# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import unittest
from typing import Optional, Union

import planet_auth
from planet_auth.auth_client import AuthClientConfig, AuthClient, AuthClientException
from planet_auth.credential import Credential
from planet_auth.oidc.auth_clients.auth_code_flow import (
    AuthCodeClientConfig,
    AuthCodeWithPubKeyClientConfig,
    AuthCodeWithClientSecretClientConfig,
)
from planet_auth.oidc.auth_clients.device_code_flow import (
    DeviceCodeClientConfig,
    DeviceCodeWithClientSecretClientConfig,
    DeviceCodeWithPubKeyClientConfig,
)
from planet_auth.oidc.auth_clients.client_credentials_flow import (
    ClientCredentialsClientSecretClientConfig,
    ClientCredentialsPubKeyClientConfig,
)
from planet_auth.oidc.auth_clients.client_validator import OidcClientValidatorAuthClientConfig
from planet_auth.oidc.auth_clients.resource_owner_flow import (
    ResourceOwnerClientConfig,
    ResourceOwnerWithClientSecretClientConfig,
    ResourceOwnerWithPubKeyClientConfig,
)
from planet_auth.planet_legacy.auth_client import PlanetLegacyAuthClientConfig
from planet_auth.request_authenticator import CredentialRequestAuthenticator
from planet_auth.static_api_key.auth_client import StaticApiKeyAuthClientConfig
from planet_auth.none.noop_auth import NoOpAuthClientConfig
from tests.test_planet_auth.util import tdata_resource_file_path


class AuthClientConfigTestImpl(AuthClientConfig):
    @classmethod
    def meta(cls):
        return {}


class AuthClientConfigInvalidImpl(AuthClientConfig):
    @classmethod
    def meta(cls):
        return {}


class AuthClientTestImpl(AuthClient):
    def __init__(self, client_config: AuthClientConfigTestImpl):
        super().__init__(client_config)
        self._test_client_config = client_config

    def login(
        self, allow_open_browser: Optional[bool] = False, allow_tty_prompt: Optional[bool] = False, **kwargs
    ) -> Credential:
        assert 0  # abstract method not under test

    def default_request_authenticator(
        self, credential: Union[pathlib.Path, Credential]
    ) -> CredentialRequestAuthenticator:
        # return SimpleInMemoryRequestAuthenticator(token_body=None)
        assert 0  # abstract method not under test

    def can_login_unattended(self) -> bool:
        return False


class TestAuthClientBase(unittest.TestCase):
    def test_no_impl_exception(self):
        under_test = AuthClientTestImpl(AuthClientConfigTestImpl())

        with self.assertRaises(AuthClientException):
            under_test.refresh(None, None)

        with self.assertRaises(AuthClientException):
            under_test.validate_access_token_remote(None)

        with self.assertRaises(AuthClientException):
            under_test.validate_access_token_local(None, None)

        with self.assertRaises(AuthClientException):
            under_test.validate_id_token_remote(None)

        with self.assertRaises(AuthClientException):
            under_test.validate_id_token_local(None)

        with self.assertRaises(AuthClientException):
            under_test.validate_refresh_token_remote(None)

        with self.assertRaises(AuthClientException):
            under_test.revoke_access_token(None)

        with self.assertRaises(AuthClientException):
            under_test.revoke_refresh_token(None)

        with self.assertRaises(AuthClientException):
            under_test.device_login_initiate()

        with self.assertRaises(AuthClientException):
            under_test.device_login_complete({})

        with self.assertRaises(AuthClientException):
            under_test.userinfo_from_access_token(None)

        with self.assertRaises(AuthClientException):
            under_test.get_scopes()


class ClientFactoryTest(unittest.TestCase):
    def test_create_pkce_auth_code_client(self):
        self.assertIsInstance(
            AuthClient.from_config(AuthCodeClientConfig(auth_server="dummy", client_id="dummy", redirect_uri="dummy")),
            planet_auth.oidc.auth_clients.auth_code_flow.AuthCodeAuthClient,
        )

    def test_create_pkce_auth_code_client_secret_client(self):
        self.assertIsInstance(
            AuthClient.from_config(
                AuthCodeWithClientSecretClientConfig(
                    auth_server="dummy", client_id="dummy", redirect_uri="dummy", client_secret="dummy"
                )
            ),
            planet_auth.oidc.auth_clients.auth_code_flow.AuthCodeWithClientSecretAuthClient,
        )

    def test_create_pkce_auth_code_pubkey_client(self):
        self.assertIsInstance(
            AuthClient.from_config(
                AuthCodeWithPubKeyClientConfig(
                    auth_server="dummy", client_id="dummy", redirect_uri="dummy", client_privkey="dummy"
                )
            ),
            planet_auth.oidc.auth_clients.auth_code_flow.AuthCodeWithPubKeyAuthClient,
        )

    def test_create_device_code_client(self):
        self.assertIsInstance(
            AuthClient.from_config(DeviceCodeClientConfig(auth_server="dummy", client_id="dummy")),
            planet_auth.oidc.auth_clients.device_code_flow.DeviceCodeAuthClient,
        )

    def test_create_device_code_client_secret_client(self):
        self.assertIsInstance(
            AuthClient.from_config(
                DeviceCodeWithClientSecretClientConfig(auth_server="dummy", client_id="dummy", client_secret="dummy")
            ),
            planet_auth.oidc.auth_clients.device_code_flow.DeviceCodeWithClientSecretAuthClient,
        )

    def test_create_device_code_pubkey_client(self):
        self.assertIsInstance(
            AuthClient.from_config(
                DeviceCodeWithPubKeyClientConfig(auth_server="dummy", client_id="dummy", client_privkey="dummy")
            ),
            planet_auth.oidc.auth_clients.device_code_flow.DeviceCodeWithPubKeyAuthClient,
        )

    def test_create_client_credentials_client_secret_client(self):
        self.assertIsInstance(
            AuthClient.from_config(
                ClientCredentialsClientSecretClientConfig(
                    auth_server="dummy", client_id="dummy", client_secret="dummy"
                )
            ),
            planet_auth.oidc.auth_clients.client_credentials_flow.ClientCredentialsClientSecretAuthClient,
        )

    def test_create_client_credentials_pubkey_client(self):
        self.assertIsInstance(
            AuthClient.from_config(
                ClientCredentialsPubKeyClientConfig(auth_server="dummy", client_id="dummy", client_privkey="dummy")
            ),
            planet_auth.oidc.auth_clients.client_credentials_flow.ClientCredentialsPubKeyAuthClient,
        )

    def test_create_resource_owner_client(self):
        self.assertIsInstance(
            AuthClient.from_config(
                ResourceOwnerClientConfig(auth_server="dummy", client_id="dummy", username="dummy", password="dummy")
            ),
            planet_auth.oidc.auth_clients.resource_owner_flow.ResourceOwnerAuthClient,
        )

    def test_create_resource_owner_with_client_secret_client(self):
        self.assertIsInstance(
            AuthClient.from_config(
                ResourceOwnerWithClientSecretClientConfig(
                    auth_server="dummy", client_id="dummy", username="dummy", password="dummy", client_secret="dummy"
                )
            ),
            planet_auth.oidc.auth_clients.resource_owner_flow.ResourceOwnerWithClientSecretAuthClient,
        )

    def test_create_resource_owner_with_pubkey_client(self):
        self.assertIsInstance(
            AuthClient.from_config(
                ResourceOwnerWithPubKeyClientConfig(
                    auth_server="dummy", client_id="dummy", username="dummy", password="dummy", client_privkey="dummy"
                )
            ),
            planet_auth.oidc.auth_clients.resource_owner_flow.ResourceOwnerWithPubKeyAuthClient,
        )

    def test_create_client_validator_client(self):
        self.assertIsInstance(
            AuthClient.from_config(OidcClientValidatorAuthClientConfig(auth_server="dummy")),
            planet_auth.oidc.auth_clients.client_validator.OidcClientValidatorAuthClient,
        )

    def test_create_planet_legacy_client(self):
        self.assertIsInstance(
            AuthClient.from_config(PlanetLegacyAuthClientConfig(legacy_auth_endpoint="dummy")),
            planet_auth.planet_legacy.auth_client.PlanetLegacyAuthClient,
        )

    def test_static_apikey_client(self):
        self.assertIsInstance(
            AuthClient.from_config(StaticApiKeyAuthClientConfig()),
            planet_auth.static_api_key.auth_client.StaticApiKeyAuthClient,
        )

    def test_noop_client(self):
        self.assertIsInstance(
            AuthClient.from_config(NoOpAuthClientConfig(extra_garbage_to_ignore="some_test_trash")),
            planet_auth.none.noop_auth.NoOpAuthClient,
        )

    def test_invalid_config_type(self):
        with self.assertRaises(AuthClientException):
            AuthClient.from_config(AuthClientConfigInvalidImpl())


class ConfigFactoryTest(unittest.TestCase):
    def test_pkce_auth_code_config_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/auth_code.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, AuthCodeClientConfig)

    def test_pkce_auth_code_secret_config_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/auth_code_secret.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, AuthCodeWithClientSecretClientConfig)

    def test_pkce_auth_code_pubkey_config_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/auth_code_pubkey.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, AuthCodeWithPubKeyClientConfig)

    def test_device_code_config_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/device_auth.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, DeviceCodeClientConfig)

    def test_device_code_secret_config_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/device_auth_secret.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, DeviceCodeWithClientSecretClientConfig)

    def test_device_code_pubkey_config_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/device_auth_pubkey.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, DeviceCodeWithPubKeyClientConfig)

    def test_client_credentials_client_secret_config_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/client_credentials_client_secret.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, ClientCredentialsClientSecretClientConfig)

    def test_client_credentials_pubkey_config_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/client_credentials_pubkey_file.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, ClientCredentialsPubKeyClientConfig)

    def test_resource_owner_config_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/resource_owner.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, ResourceOwnerClientConfig)

    def test_resource_owner_client_secret_config_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/resource_owner_secret.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, ResourceOwnerWithClientSecretClientConfig)

    def test_resource_owner_pubkey_config_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/resource_owner_pubkey.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, ResourceOwnerWithPubKeyClientConfig)

    def test_client_validator_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/client_validator.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, OidcClientValidatorAuthClientConfig)

    def test_static_config_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/static_api_key.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, StaticApiKeyAuthClientConfig)

    def test_noop_config_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/none.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, NoOpAuthClientConfig)

    def test_planet_legacy_config_from_file(self):
        file_path = tdata_resource_file_path("auth_client_configs/utest/planet_legacy.json")
        auth_client_config = AuthClientConfig.from_file(file_path)
        self.assertIsInstance(auth_client_config, PlanetLegacyAuthClientConfig)

    def test_invalid_config_type(self):
        with self.assertRaises(AuthClientException):
            AuthClientConfig.from_dict({"client_type": "__test_invalid__"})


class ValidateConfigMeta(unittest.TestCase):
    """
    Test config metadata used by the Auth classes themselves
    """

    def _check_meta_key_uniqueness(self, under_test_meta_key):
        under_test_meta_values = {}
        for under_test_config_type in AuthClientConfig._get_config_types():
            under_test_meta = under_test_config_type.meta()
            under_test_meta_value = under_test_meta.get(under_test_meta_key)
            self.assertIsNotNone(
                under_test_meta.get(under_test_meta_key),
                "Meta field {} is not set for type {}".format(under_test_meta_key, under_test_config_type),
            )
            self.assertIsNone(
                under_test_meta_values.get(under_test_meta.get(under_test_meta_key)),
                "Meta field {} must be unique across auth client config types.  Duplicate value '{}' found in '{}'".format(
                    under_test_meta_key, under_test_meta_value, under_test_config_type.__name__
                ),
            )
            under_test_meta_values[under_test_meta_value] = under_test_config_type.__name__

    def test_type_name_uniqueness(self):
        self._check_meta_key_uniqueness(AuthClientConfig.CLIENT_TYPE_KEY)

    def test_client_type_uniqueness(self):
        self._check_meta_key_uniqueness("auth_client_class")


class ValidateConfigMetaForCLI(unittest.TestCase):
    """
    Test config metadata used the the auth client CLI
    """

    def test_hints_are_well_formed(self):
        for under_test_config_type in AuthClientConfig._get_config_types():
            under_tests_hints = under_test_config_type.meta().get("config_hints")
            if under_tests_hints:
                for under_test_hint in under_tests_hints:
                    self.assertIsInstance(
                        under_test_hint,
                        dict,
                        "config_hints for type {} must be dict type".format(under_test_config_type.__name__),
                    )
                    # we check for truthy rather than not being none since empty strings are not really useful either
                    self.assertTrue(
                        under_test_hint.get("config_key"),
                        "Meta {} for {} missing a value for {}".format(
                            "config_hint", under_test_config_type.__name__, "config_key"
                        ),
                    )
                    self.assertTrue(
                        under_test_hint.get("config_key_name"),
                        "Meta {} for {} missing a value for {}".format(
                            "config_hint", under_test_config_type.__name__, "config_key_name"
                        ),
                    )
                    self.assertTrue(
                        under_test_hint.get("config_key_description"),
                        "Meta {} for {} missing a value for {}".format(
                            "config_hint", under_test_config_type.__name__, "config_key_description"
                        ),
                    )

    def test_no_duplicate_hints(self):
        for under_test_config_type in AuthClientConfig._get_config_types():
            under_tests_hints = under_test_config_type.meta().get("config_hints")
            if under_tests_hints:
                conf_hints = {}
                for under_test_hint in under_tests_hints:
                    self.assertIsNone(
                        conf_hints.get(under_test_hint.get("config_key")),
                        "Duplicate configuration hint for the key '{}' found in '{}'".format(
                            under_test_hint.get("config_key"), under_test_config_type.__name__
                        ),
                    )
                    conf_hints[under_test_hint.get("config_key")] = 1



import http
import logging
import os
import re
from flask import Flask, make_response, request
from functools import wraps
from typing import List, Optional

import planet_auth
import planet_auth_config  # type: ignore

#############################################################################
# Logging Configuration
#############################################################################
# In a simple environment, the global logging format will be
# applied. This example removes all formatting around the log message from
# the outer application, allowing the library's logging to be emitted with
# no further decoration.  The auth library will emit logs in a json
# format, so removing surrounding formatting insures that the json will be
# parsable:
logging.basicConfig(format="%(message)s", level=logging.DEBUG)

# Alternatively, we can tell the library to use python structured logging.
# The exact appearance of the logs in this case will be dependent on the
# configuration of the logger.  When structured loging is used, the library
# will log using the python logger's `extra` data.  In most cases,
# this would be expected to be done in conjunction with a call to
# setPyLoggerForAuthLogger() (see below).
#
#     planet_auth.setStructuredLogging()

# What logger will be used can also be controlled.  By default, the library
# will use a logger it creates internally that is global to the library.
# Applications may set the logger for the library to "None" to quite all,
# or may set it to any logger of their choice.:
#
#     planet_auth.setPyLoggerForAuthLogger(None)
#   or
#     planet_auth.setPyLoggerForAuthLogger(logging.getLogger(name="my-custom-auth-lib-logger"))


#############################################################################
# Validator Configuration
#############################################################################
# The Planet auth config package provides constants that describe
# various Planet environments.  The environments described in that
# package are not considered part of the auth library, which is
# deployment agnostic, and serves both client and server use cases.
#
# Services may lean on the config library to provide easy deployment
# ergonomics for common use cases, and benefit from published config
# value changes.  But, services should also maintain the ability
# to be deployed to arbitrary trust environments for development
# and testing purposes.
#
# Never ever ever EVER NEVER EVER accept staging authorities in a
# production deployment.
# NEVER. NEVER NEVER.
# Do not cross the streams.
# Seriously.  Don't do it.

named_env = os.getenv("MY_DEPLOYMENT_ENV")
if named_env == "production":
    auth_validator = planet_auth.OidcMultiIssuerValidator.from_auth_server_configs(
        trusted_auth_server_configs=[planet_auth_config.Production.PRIMARY_PUBLIC_OAUTH_AUTHORITY],
        secondary_auth_server_configs=[planet_auth_config.Production.DEPRECATED_PUBLIC_OAUTH_AUTHORITY], # type: ignore
    )
elif named_env == "staging":
    auth_validator = planet_auth.OidcMultiIssuerValidator.from_auth_server_configs(
        trusted_auth_server_configs=[planet_auth_config.Staging.PRIMARY_PUBLIC_OAUTH_AUTHORITY],
        secondary_auth_server_configs=[planet_auth_config.Staging.DEPRECATED_PUBLIC_OAUTH_AUTHORITY], # type: ignore
    )
else:
    # in this example, MY_DEPLOYMENT_ENV trumps any other settings
    auth_validator = planet_auth.OidcMultiIssuerValidator.from_auth_server_configs(
        trusted_auth_server_configs=[
            {
                "auth_server": os.getenv("MY_TRUSTED_ISSUER_PRIMARY"),
                "audiences": [os.getenv("MY_TRUSTED_AUDIENCE_PRIMARY")],
            },
            {
                "auth_server": os.getenv("MY_TRUSTED_ISSUER_DEPRECATED"),
                "audiences": [os.getenv("MY_TRUSTED_AUDIENCE_DEPRECATED")],
            }
        ],
    )


#############################################################################
# Connecting validator to flask using decorators (your service needs may vary)
#############################################################################
# Note: This example only enforces basic authentication - that the user has
#       authenticated to a trusted auth server and presented a valid access
#       token.  In practice, services are also responsible for determining
#       whether that particular client user is allowed to make a particular
#       request.  Services may have their own database or configuration that
#       governs this. The overall service architecture may define token
#       claims that convey particular privileges. Neither of these are
#       demonstrated by this example.
@planet_auth.AuthException.recast(ValueError)
def flask_auth_worker(do_remote_check: bool, scopes_any_of=Optional[List[str]]):
    auth_header = request.headers.get("Authorization")
    if not auth_header:
        raise planet_auth.AuthException("Authorization header not provided")

    scheme, token = re.split(" +", auth_header.strip(), 1)
    if scheme != "Bearer":
        raise planet_auth.AuthException("Unrecognized authentication token scheme: {}".format(scheme))

    # this call will throw auth failures
    return auth_validator.validate_access_token(
        token=token, do_remote_revocation_check=do_remote_check, scopes_anyof=scopes_any_of
    )


def flask_auth_decorator(do_remote_check: bool = False, scopes_any_of=None):
    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kws):
            try:
                # This example does not pass the results of the
                # validation back to the flask app at this time.
                local_validation, remote_validation = flask_auth_worker(
                    do_remote_check=do_remote_check, scopes_any_of=scopes_any_of
                )
                # If you have additional application specific checks
                # you wish to make at the granularity of the decorator,
                # this is a reasonable place to do them. After having
                # passed the checks above, the contents of the token
                # as returned in local_validation / remote_validation
                # can be considered trusted.
            except planet_auth.AuthException as ae:
                # Don't leak information to unauthorized clients in the response.
                # It's fine to be more detailed in server side logs.
                logging.warning("Auth failure: {}".format(ae))
                # Return or abort, as per your application conventions.
                return make_response("UNAUTHORIZED", http.HTTPStatus.UNAUTHORIZED)
                # abort(http.HTTPStatus.UNAUTHORIZED, "UNAUTHORIZED")
            return fn(*args, **kws)

        return wrapper

    return decorator


#############################################################################
# Main flask application
#############################################################################
app = Flask(__name__)


# Order of decorators matters
@app.route("/require-scopes")
@flask_auth_decorator(scopes_any_of=["required_scope1", "required_scope2"])
def require_scopes():
    return "Hello World! from a scope restricted endpoint.\n"


@app.route("/")
@flask_auth_decorator(scopes_any_of=None)
def hello():
    return "Hello World!\n"


if __name__ == "__main__":
    # Control how the library emits its logs:
    # planet_auth.setStringLogging()
    # planet_auth.setStructuredLogging()
    app.run(host="localhost", port=5001)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

from requests.models import Response
from unittest import mock

from planet_auth.oidc.api_clients.discovery_api_client import DiscoveryApiClient

TEST_API_ENDPOINT = "https://blackhole.unittest.planet.com/api"
API_RESPONSE_VALID = {
    "issuer": "https://blackhole.unittest.planet.com/oauth2",
    "authorization_endpoint": "https://blackhole.unittest.planet.com/oauth2/v1/authorize",
    "token_endpoint": "https://blackhole.unittest.planet.com/oauth2/v1/token",
    "userinfo_endpoint": "https://blackhole.unittest.planet.com/oauth2/v1/userinfo",
    "registration_endpoint": "https://account.planet.com/oauth2/v1/clients",
    "jwks_uri": "https://blackhole.unittest.planet.com/oauth2/v1/keys",
    "response_types_supported": [
        "code",
        "id_token",
        "code id_token",
        "code token",
        "id_token token",
        "code id_token token",
    ],
    "response_modes_supported": ["query", "fragment", "form_post", "okta_post_message"],
    "grant_types_supported": [
        "authorization_code",
        "implicit",
        "refresh_token",
        "password",
        "urn:ietf:params:oauth:grant-type:device_code",
    ],
    "subject_types_supported": ["public"],
    "id_token_signing_alg_values_supported": ["RS256"],
    "scopes_supported": ["openid", "profile", "email", "address", "phone", "offline_access", "device_sso"],
    "token_endpoint_auth_methods_supported": [
        "client_secret_basic",
        "client_secret_post",
        "client_secret_jwt",
        "private_key_jwt",
        "none",
    ],
    "claims_supported": [
        "iss",
        "ver",
        "sub",
        "aud",
        "iat",
        "exp",
        "jti",
        "auth_time",
        "amr",
        "idp",
        "nonce",
        "name",
        "nickname",
        "preferred_username",
        "given_name",
        "middle_name",
        "family_name",
        "email",
        "email_verified",
        "profile",
        "zoneinfo",
        "locale",
        "address",
        "phone_number",
        "picture",
        "website",
        "gender",
        "birthdate",
        "updated_at",
        "at_hash",
        "c_hash",
    ],
    "code_challenge_methods_supported": ["S256"],
    "introspection_endpoint": "https://blackhole.unittest.planet.com/oauth2/v1/introspect",
    "introspection_endpoint_auth_methods_supported": [
        "client_secret_basic",
        "client_secret_post",
        "client_secret_jwt",
        "private_key_jwt",
        "none",
    ],
    "revocation_endpoint": "https://blackhole.unittest.planet.com/oauth2/v1/revoke",
    "revocation_endpoint_auth_methods_supported": [
        "client_secret_basic",
        "client_secret_post",
        "client_secret_jwt",
        "private_key_jwt",
        "none",
    ],
    "end_session_endpoint": "https://blackhole.unittest.planet.com/oauth2/v1/logout",
    "request_parameter_supported": True,
    "request_object_signing_alg_values_supported": [
        "HS256",
        "HS384",
        "HS512",
        "RS256",
        "RS384",
        "RS512",
        "ES256",
        "ES384",
        "ES512",
    ],
    "device_authorization_endpoint": "https://blackhole.unittest.planet.com/oauth2/v1/device/authorize",
}
# API_RESPONSE_FAILED = {}


class MockRequests:
    request_counter = 0

    @staticmethod
    def mocked_response_ok(request_url, **kwargs):
        MockRequests.request_counter += 1
        response = Response()
        response.status_code = 200
        response.headers["content-type"] = "application/json"
        response._content = str.encode(json.dumps(API_RESPONSE_VALID))
        return response


class DiscoveryApiClientTest(unittest.TestCase):
    @mock.patch("requests.sessions.Session.get", side_effect=MockRequests.mocked_response_ok)
    def test_verify_caching(self, mock_get):
        self.assertEqual(0, MockRequests.request_counter)
        under_test = DiscoveryApiClient(discovery_uri=TEST_API_ENDPOINT)
        json_response = under_test.discovery()
        self.assertEqual(API_RESPONSE_VALID, json_response)
        self.assertEqual(1, MockRequests.request_counter)
        under_test.discovery()
        self.assertEqual(1, MockRequests.request_counter)

    def test_discovery_url_fixups(self):
        under_test = DiscoveryApiClient(discovery_uri="/test_discover_uri")
        self.assertEqual("/test_discover_uri", under_test._endpoint_uri)

        under_test = DiscoveryApiClient(auth_server="/test_auth_server_base")
        self.assertEqual("/test_auth_server_base/.well-known/openid-configuration", under_test._endpoint_uri)

        under_test = DiscoveryApiClient(auth_server="/test_auth_server_base/")
        self.assertEqual("/test_auth_server_base/.well-known/openid-configuration", under_test._endpoint_uri)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

from requests.auth import AuthBase
from requests.models import Response
from typing import Tuple, Optional
from unittest import mock

from planet_auth.oidc.api_clients.revocation_api_client import RevocationApiClient

TEST_API_ENDPOINT = "https://blackhole.unittest.planet.com/api"
TEST_ACCESS_TOKEN = "__test_access_token__"
TEST_REFRESH_TOKEN = "__test_refresh_token__"
API_RESPONSE_VALID = {}
API_RESPONSE_FAILED = {}


def noop_auth_enricher(raw_payload: dict, audience: str) -> Tuple[dict, Optional[AuthBase]]:
    return raw_payload, None


def mocked_response_ok(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(API_RESPONSE_VALID))
    return response


class RevocationApiClientTest(unittest.TestCase):
    """
    Revocation API testing is very boring.  By design, the endpoint doesn't
    tell you much. All of our expected error handling is tested with the base
    class.
    """

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_ok)
    def test_revoke_access_token_with_enricher(self, mock_post):
        under_test = RevocationApiClient(revocation_uri=TEST_API_ENDPOINT)
        under_test.revoke_access_token(TEST_ACCESS_TOKEN, noop_auth_enricher)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_ok)
    def test_revoke_access_token_without_enricher(self, mock_post):
        under_test = RevocationApiClient(revocation_uri=TEST_API_ENDPOINT)
        under_test.revoke_access_token(TEST_ACCESS_TOKEN, None)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_ok)
    def test_revoke_refresh_token(self, mock_post):
        under_test = RevocationApiClient(revocation_uri=TEST_API_ENDPOINT)
        under_test.revoke_refresh_token(TEST_REFRESH_TOKEN, None)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from typing import Tuple

from planet_auth.auth_client import AuthClientConfig
from planet_auth.oidc.auth_client import OidcAuthClient
from planet_auth.oidc.oidc_credential import FileBackedOidcCredential
from planet_auth.oidc.api_clients.introspect_api_client import IntrospectionApiException
from tests.test_planet_auth.util import tdata_resource_file_path

from planet_auth.auth import Auth

CLIENT_CONFIG_FOR_USER_LOGIN = [
    "oauth_pkce_auth_code.json",
    "oauth_pkce_auth_code_client_secret.sops.json",
    "oauth_pkce_auth_code_pubkey.sops.json",
    "oauth_resource_owner.sops.json",
    "oauth_resource_owner_client_secret.sops.json",
    "oauth_resource_owner_client_pubkey.sops.json",
]

CLIENT_CONIGS_FOR_CLIENT_LOGIN = [
    "oauth_client_credentials_client_secret.sops.json",
    "oauth_client_credentials_pubkey_literal.sops.json",
]

## from tests.util import is_not_interactive_shell
## @pytest.mark.skipif(condition=is_not_interactive_shell(), reason="Skipping test in non-interactive shell")


#
# WARNING:
#    These test do not run in parallel well.  There is contention over callback the listening port.
#    These tests may also require user interaction.  These are here to make it convenient to run
#    a battery of tests manually against a wide swath of client configs.  But, these are not
#    at this point fire and forget tests.
#
class TestLiveService:
    @staticmethod
    def auth_from_test_resource_file(resource_file) -> Tuple[Auth, OidcAuthClient, FileBackedOidcCredential]:
        test_client_conf_file = tdata_resource_file_path("auth_client_configs/live_service/{}".format(resource_file))
        auth = Auth.initialize_from_config(client_config=AuthClientConfig.from_file(test_client_conf_file))
        oidc_auth_client = auth.auth_client()
        token = oidc_auth_client.login()
        return auth, oidc_auth_client, token

    @staticmethod
    def revoke_test_token(auth_client: OidcAuthClient, token: FileBackedOidcCredential):
        if token:
            if token.refresh_token():
                auth_client.revoke_refresh_token(token.refresh_token())
            if token.access_token():
                auth_client.revoke_access_token(token.access_token())

    @pytest.mark.parametrize("client_config_file", CLIENT_CONFIG_FOR_USER_LOGIN)
    def test_user_login(self, client_config_file):
        """
        Test OIDC flows that login for a user
        """
        auth_context, oidc_client_under_test, token = self.auth_from_test_resource_file(client_config_file)
        try:
            assert token.access_token() is not None
            assert token.id_token() is not None
            assert token.refresh_token() is not None
        finally:
            self.revoke_test_token(auth_client=oidc_client_under_test, token=token)

    @pytest.mark.parametrize("client_config_file", CLIENT_CONIGS_FOR_CLIENT_LOGIN)
    def test_client_login(self, client_config_file):
        """
        Test OIDC flows that login for the client itself
        """
        auth_context, oidc_client_under_test, token = self.auth_from_test_resource_file(client_config_file)
        try:
            assert token.access_token() is not None
            assert token.id_token() is None
            assert token.refresh_token() is None
        finally:
            self.revoke_test_token(auth_client=oidc_client_under_test, token=token)

    @pytest.mark.parametrize("client_config_file", CLIENT_CONFIG_FOR_USER_LOGIN)
    def test_validate_id_token(self, client_config_file):
        auth_context, oidc_client_under_test, token = self.auth_from_test_resource_file(client_config_file)
        try:
            validation = oidc_client_under_test.validate_id_token_remote(token.id_token())
            assert isinstance(validation.get("active"), bool)
            assert validation.get("active")
        finally:
            self.revoke_test_token(auth_client=oidc_client_under_test, token=token)

    @pytest.mark.parametrize("client_config_file", CLIENT_CONFIG_FOR_USER_LOGIN + CLIENT_CONIGS_FOR_CLIENT_LOGIN)
    def test_validate_access_token(self, client_config_file):
        auth_context, oidc_client_under_test, token = self.auth_from_test_resource_file(client_config_file)
        try:
            validation = oidc_client_under_test.validate_access_token_remote(token.access_token())
            assert isinstance(validation.get("active"), bool)
            assert validation.get("active")
        finally:
            self.revoke_test_token(auth_client=oidc_client_under_test, token=token)

    @pytest.mark.parametrize("client_config_file", CLIENT_CONFIG_FOR_USER_LOGIN + CLIENT_CONIGS_FOR_CLIENT_LOGIN)
    def test_revoke_access_token(self, client_config_file):
        auth_context, oidc_client_under_test, token = self.auth_from_test_resource_file(client_config_file)
        try:
            oidc_client_under_test.revoke_access_token(token.access_token())
            with pytest.raises(IntrospectionApiException):
                oidc_client_under_test.validate_access_token_remote(token.access_token())
        finally:
            self.revoke_test_token(auth_client=oidc_client_under_test, token=token)

    @pytest.mark.parametrize("client_config_file", CLIENT_CONFIG_FOR_USER_LOGIN)
    def test_validate_refresh_token(self, client_config_file):
        auth_context, oidc_client_under_test, token = self.auth_from_test_resource_file(client_config_file)
        try:
            validation = oidc_client_under_test.validate_refresh_token_remote(token.refresh_token())
            assert isinstance(validation.get("active"), bool)
            assert validation.get("active")
        finally:
            self.revoke_test_token(auth_client=oidc_client_under_test, token=token)

    @pytest.mark.parametrize("client_config_file", CLIENT_CONFIG_FOR_USER_LOGIN)
    def test_revoke_refresh_token(self, client_config_file):
        auth_context, oidc_client_under_test, token = self.auth_from_test_resource_file(client_config_file)
        try:
            oidc_client_under_test.revoke_refresh_token(token.refresh_token())
            with pytest.raises(IntrospectionApiException):
                oidc_client_under_test.validate_refresh_token_remote(token.refresh_token())
        finally:
            self.revoke_test_token(auth_client=oidc_client_under_test, token=token)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The Planet Authentication package

This package contains functionality for authenticating clients to the
service and managing authentication material.  This package knows nothing
about the service itself apart from how to interact with authentication
APIs.

This package understands multiple authentication mechanisms, whose details
are encapsulated in implementation subclasses that implement the primary
(mostly abstract) base class interfaces.

With many different ways to interact with authentication services and
ways to use the resulting authorization material (OAuth tokens,
API keys, etc), configuration is important to understand.
See the documentation for [configuration](../configuration).

The primary interfaces implemented for users of this package are as follows:

- [planet_auth.Auth][] - A container class for initializing and grouping
      a working set of authentication objects (below).
- [planet_auth.AuthClient][] & [planet_auth.AuthClientConfig][] - Responsible for
      interacting with authentication services to obtain a credential that
      may be used with other API requests. Different clients have different
      configuration needs, so a configuration type exists for each client
      type to keep configuration on rails.
- [planet_auth.Credential][] - Models just a credential.
      Responsible for reading and writing saved credentials to disk and
      performing basic data validation.  Knows nothing about how to get a
      credential, or how to use a credential.
- [planet_auth.RequestAuthenticator][] - Responsible for
      decorating API requests with a credential. Compatible with `httpx` and
      `requests` libraries.  Some authentication mechanisms require that
      the request authenticator also have an
      [AuthClient][planet_auth.AuthClient], others do not.  Whether or not
      this is required is driven by the specifics of the authentication
      mechanism.
"""

from .auth import Auth
from .auth_exception import AuthException
from .auth_client import AuthClientConfig, AuthClient
from .credential import Credential
from .request_authenticator import RequestAuthenticator, CredentialRequestAuthenticator
from .logging.auth_logger import setPyLoggerForAuthLogger, setStringLogging, setStructuredLogging

from .oidc.auth_client import OidcAuthClient, OidcAuthClientConfig
from .oidc.auth_clients.auth_code_flow import (
    AuthCodeClientConfig,
    AuthCodeAuthClient,
    AuthCodeWithClientSecretClientConfig,
    AuthCodeWithClientSecretAuthClient,
    AuthCodeWithPubKeyClientConfig,
    AuthCodeWithPubKeyAuthClient,
    AuthCodeAuthClientException,
)
from .oidc.auth_clients.client_credentials_flow import (
    ClientCredentialsAuthClientBase,
    ClientCredentialsClientSecretClientConfig,
    ClientCredentialsClientSecretAuthClient,
    ClientCredentialsPubKeyClientConfig,
    ClientCredentialsPubKeyAuthClient,
)
from .oidc.auth_clients.device_code_flow import (
    DeviceCodeClientConfig,
    DeviceCodeAuthClient,
    DeviceCodeWithClientSecretClientConfig,
    DeviceCodeWithClientSecretAuthClient,
    DeviceCodeWithPubKeyClientConfig,
    DeviceCodeWithPubKeyAuthClient,
    DeviceCodeAuthClientException,
)
from .oidc.auth_clients.client_validator import (
    OidcClientValidatorAuthClientConfig,
    OidcClientValidatorAuthClient,
)
from .oidc.auth_clients.resource_owner_flow import (
    ResourceOwnerClientConfig,
    ResourceOwnerAuthClient,
    ResourceOwnerWithClientSecretClientConfig,
    ResourceOwnerWithClientSecretAuthClient,
    ResourceOwnerWithPubKeyClientConfig,
    ResourceOwnerWithPubKeyAuthClient,
    ResourceOwnerAuthClientException,
)
from .oidc.token_validator import (
    ExpiredTokenException,
    InvalidAlgorithmTokenException,
    InvalidArgumentException,
    InvalidTokenException,
    ScopeNotGrantedTokenException,
    TokenValidator,
    TokenValidatorException,
    UnknownSigningKeyTokenException,
)
from .oidc.multi_validator import OidcMultiIssuerValidator
from .planet_legacy.auth_client import PlanetLegacyAuthClientConfig, PlanetLegacyAuthClient
from .static_api_key.auth_client import (
    StaticApiKeyAuthClientConfig,
    StaticApiKeyAuthClient,
    StaticApiKeyAuthClientException,
)
from .none.noop_auth import NoOpAuthClientConfig, NoOpAuthClient

from .oidc.oidc_credential import FileBackedOidcCredential
from .planet_legacy.legacy_api_key import FileBackedPlanetLegacyApiKey
from .static_api_key.static_api_key import FileBackedApiKey

from .oidc.request_authenticator import (
    RefreshingOidcTokenRequestAuthenticator,
    RefreshOrReloginOidcTokenRequestAuthenticator,
)
from .planet_legacy.request_authenticator import PlanetLegacyRequestAuthenticator
from .static_api_key.request_authenticator import FileBackedApiKeyRequestAuthenticator

from .storage_utils import (
    FileBackedJsonObject,
    FileBackedJsonObjectException,
    InvalidDataException,
    ObjectStorageProvider,
    ObjectStorageProvider_KeyType,
)

__all__ = [
    # Classes
    "Auth",
    "AuthClient",
    "AuthClientConfig",
    "AuthCodeAuthClient",
    "AuthCodeClientConfig",
    "AuthCodeWithClientSecretAuthClient",
    "AuthCodeWithClientSecretClientConfig",
    "AuthCodeWithPubKeyAuthClient",
    "AuthCodeWithPubKeyClientConfig",
    "AuthCodeAuthClientException",
    "AuthException",
    "ClientCredentialsAuthClientBase",
    "ClientCredentialsClientSecretAuthClient",
    "ClientCredentialsClientSecretClientConfig",
    "ClientCredentialsPubKeyAuthClient",
    "ClientCredentialsPubKeyClientConfig",
    "CredentialRequestAuthenticator",
    "DeviceCodeClientConfig",
    "DeviceCodeAuthClient",
    "DeviceCodeWithClientSecretClientConfig",
    "DeviceCodeWithClientSecretAuthClient",
    "DeviceCodeWithPubKeyClientConfig",
    "DeviceCodeWithPubKeyAuthClient",
    "DeviceCodeAuthClientException",
    "Credential",
    "ExpiredTokenException",
    "FileBackedJsonObject",
    "FileBackedJsonObjectException",
    "FileBackedApiKey",
    "FileBackedApiKeyRequestAuthenticator",
    "FileBackedOidcCredential",
    "FileBackedPlanetLegacyApiKey",
    "InvalidAlgorithmTokenException",
    "InvalidDataException",
    "InvalidArgumentException",
    "InvalidTokenException",
    "ObjectStorageProvider",
    "ObjectStorageProvider_KeyType",
    "OidcAuthClient",
    "OidcAuthClientConfig",
    "OidcClientValidatorAuthClientConfig",
    "OidcClientValidatorAuthClient",
    "OidcMultiIssuerValidator",
    "NoOpAuthClient",
    "NoOpAuthClientConfig",
    "PlanetLegacyAuthClient",
    "PlanetLegacyAuthClientConfig",
    "PlanetLegacyRequestAuthenticator",
    "RefreshOrReloginOidcTokenRequestAuthenticator",
    "RefreshingOidcTokenRequestAuthenticator",
    "RequestAuthenticator",
    "ResourceOwnerAuthClient",
    "ResourceOwnerClientConfig",
    "ResourceOwnerWithClientSecretAuthClient",
    "ResourceOwnerWithClientSecretClientConfig",
    "ResourceOwnerWithPubKeyAuthClient",
    "ResourceOwnerWithPubKeyClientConfig",
    "ResourceOwnerAuthClientException",
    "ScopeNotGrantedTokenException",
    "StaticApiKeyAuthClient",
    "StaticApiKeyAuthClientConfig",
    "StaticApiKeyAuthClientException",
    "TokenValidator",
    "TokenValidatorException",
    "UnknownSigningKeyTokenException",
    # Functions
    "setPyLoggerForAuthLogger",
    "setStringLogging",
    "setStructuredLogging",
]



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from planet_auth.oidc.oidc_credential import FileBackedOidcCredential
from planet_auth.storage_utils import FileBackedJsonObjectException
from tests.test_planet_auth.util import tdata_resource_file_path


class TestOidcCredential(unittest.TestCase):
    def test_asserts_valid(self):
        under_test = FileBackedOidcCredential(
            data=None, credential_file=tdata_resource_file_path("keys/oidc_test_credential.json")
        )
        under_test.load()
        self.assertIsNotNone(under_test.data())

        with self.assertRaises(FileBackedJsonObjectException):
            under_test.set_data(None)

        with self.assertRaises(FileBackedJsonObjectException):
            under_test.set_data({"test": "missing all required fields"})

    def test_getters(self):
        under_test = FileBackedOidcCredential(
            data=None, credential_file=tdata_resource_file_path("keys/oidc_test_credential.json")
        )
        under_test.load()
        self.assertEqual("_dummy_access_token_", under_test.access_token())
        self.assertEqual("_dummy_refresh_token_", under_test.refresh_token())
        self.assertEqual("_dummy_id_token_", under_test.id_token())



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import Optional, Union

from planet_auth.auth_client import AuthClientConfig, AuthClient
from planet_auth.credential import Credential
from planet_auth.request_authenticator import CredentialRequestAuthenticator, SimpleInMemoryRequestAuthenticator


class NoOpCredential(Credential):
    def load(self):
        pass

    def save(self):
        pass


class NoOpAuthClientConfig(AuthClientConfig):
    """
    Auth client that does nothing.
    """

    @classmethod
    def meta(cls):
        return {
            "client_type": "none",
            "auth_client_class": NoOpAuthClient,
            "display_name": "NoOp Auth",
            "description": "Do not perform any authentication.",
            "config_hints": [],
        }


class NoOpAuthClient(AuthClient):
    def __init__(self, client_config: NoOpAuthClientConfig):
        super().__init__(client_config)
        self._noop_client_config = client_config

    def login(
        self, allow_open_browser: Optional[bool] = False, allow_tty_prompt: Optional[bool] = False, **kwargs
    ) -> Credential:
        return NoOpCredential()

    def default_request_authenticator(
        self, credential: Union[pathlib.Path, Credential]
    ) -> CredentialRequestAuthenticator:
        return SimpleInMemoryRequestAuthenticator(token_body=None)

    def refresh(self, refresh_token, requested_scopes) -> Credential:
        return self.login()

    def validate_access_token_remote(self, access_token: str) -> dict:
        return {}

    def validate_access_token_local(
        self, access_token: str, required_audience: str = None, scopes_anyof: list = None
    ) -> dict:
        return {}

    def validate_id_token_remote(self, id_token: str) -> dict:
        return {}

    def validate_id_token_local(self, id_token: str) -> dict:
        return {}

    def validate_refresh_token_remote(self, refresh_token: str) -> dict:
        return {}

    def revoke_access_token(self, access_token: str):
        pass

    def revoke_refresh_token(self, refresh_token: str):
        pass

    def get_scopes(self):
        return []

    def can_login_unattended(self) -> bool:
        return True



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import http.server
import importlib.resources as resources
import logging
import pytest
import requests
import unittest
import urllib.parse

from unittest import mock
from urllib.parse import urlparse, parse_qs

from planet_auth.oidc.api_clients.authorization_api_client import (
    AuthorizationApiClient,
    AuthorizationApiException,
    _parse_authcode_from_callback,
)
from planet_auth.oidc.util import create_pkce_challenge_verifier_pair
from tests.test_planet_auth.util import background, find_free_port, is_cicd

REQUEST_TIMEOUT = 100
TEST_API_ENDPOINT = "https://blackhole.unittest.planet.com/api"
TEST_CLIENT_ID = "_client_id_"
TEST_REDIRECT_URI_TEMPLATE = "http://localhost:{}/utest_callback_uri"
TEST_REQUESTED_SCOPES = ["scope1", "scope2"]
TEST_REQUESTED_AUDIENCES = ["aud1"]
MOCK_AUTHCODE = "_mock_authcode_"
API_RESPONSE_VALID = {}
API_RESPONSE_FAILED = {}
TEST_VERIFIER, TEST_CHALLENGE = create_pkce_challenge_verifier_pair()

logger = logging.getLogger(__name__)


def mocked_get_password(**kwargs):
    return MOCK_AUTHCODE


def mocked_generate_nonce(length):
    # Need to defeat the cryptographic nonce for some tests.
    nonce_charset = "1234567890abcdefghijklmnopqrstuvwxyz"
    return "".join([str(nonce_charset[i % 36]) for i in range(length)])


@background
def fire_and_forget_http_get(url):
    requests.get(url, timeout=REQUEST_TIMEOUT)


def mocked_browser_authserver(request_url, arg2, **kwargs):
    # TODO: assert it's constructed the way we want? The only real
    #       branching in the code under test is handling of scopes
    #       At some point, validating this URL is partially implementing
    #       an authorization server
    parsed_query_string = parse_qs(urlparse(request_url).query)
    redirect_uri = parsed_query_string.get("redirect_uri")[0]
    state = parsed_query_string.get("state")[0]
    encoded_params = urllib.parse.urlencode({"state": state, "code": MOCK_AUTHCODE})
    callback_uri = "{}?{}".format(redirect_uri, encoded_params)

    # I don't care about the result, so run in the background and forget.
    # Against a live auth service, this callback would be called by the
    # browser after an exchange with th auth server and a final redirect.
    # Mocking the browser mocks the auth server interaction
    fire_and_forget_http_get(callback_uri)


class CallbackHandlerTest(unittest.TestCase):
    # There is really no branching logic in this class other than a logging
    # handler. It's purpose in life is to recapture the flow from the browser
    # and save that data for all the other code, which should already have
    # test coverage elsewhere.
    pass


class AuthcodeCallbackParserTest(unittest.TestCase):
    def setUp(self):
        self.under_test = _parse_authcode_from_callback
        self.dummy_callback_baseurl = TEST_REDIRECT_URI_TEMPLATE.format(8080)

    def test_empty_request_throws(self):
        with self.assertRaises(AuthorizationApiException):
            self.under_test(None, None)

        with self.assertRaises(AuthorizationApiException):
            self.under_test("", None)

    def test_explicit_error_throws(self):
        encoded_params = urllib.parse.urlencode({"error": "test_error_1"})
        callback_uri = "{}?{}".format(self.dummy_callback_baseurl, encoded_params)
        with self.assertRaises(AuthorizationApiException):
            self.under_test(callback_uri, None)

        # Mix it up. We want a failure whenever there is an error, even if
        # there is also an auth code (this really shouldn't ever happen).
        encoded_params = urllib.parse.urlencode(
            {
                "error": "test_error_2",
                "error_description": "test error description",
                "code": MOCK_AUTHCODE,
                "state": mocked_generate_nonce(8),
            }
        )
        callback_uri = "{}?{}".format(self.dummy_callback_baseurl, encoded_params)
        with self.assertRaises(AuthorizationApiException):
            self.under_test(callback_uri, None)

    # @mock.patch('planet_auth.oidc.util.generate_nonce', mocked_generate_nonce)
    def test_state_is_checked(self):
        test_state_1 = mocked_generate_nonce(8)
        test_state_2 = test_state_1 + "_STATE_IS_CORRUPTED"

        encoded_params = urllib.parse.urlencode({"code": MOCK_AUTHCODE, "state": test_state_1})
        callback_uri = "{}?{}".format(self.dummy_callback_baseurl, encoded_params)
        auth_code = self.under_test(callback_uri, test_state_1)
        self.assertEqual(MOCK_AUTHCODE, auth_code)

        encoded_params = urllib.parse.urlencode({"code": MOCK_AUTHCODE, "state": test_state_2})
        callback_uri = "{}?{}".format(self.dummy_callback_baseurl, encoded_params)
        with self.assertRaises(AuthorizationApiException):
            self.under_test(callback_uri, test_state_1)

    def test_callback_not_understood_throws(self):
        encoded_params = urllib.parse.urlencode({"data1": "some random data"})
        callback_uri = "{}?{}".format(self.dummy_callback_baseurl, encoded_params)
        with self.assertRaises(AuthorizationApiException):
            self.under_test(callback_uri, None)


class AuthorizationApiClientTest(unittest.TestCase):
    def setUp(self):
        self.callback_port = find_free_port()
        self.callback_uri = TEST_REDIRECT_URI_TEMPLATE.format(self.callback_port)
        self.pkce_verifier, self.pkce_challenge = create_pkce_challenge_verifier_pair()

    @pytest.mark.skipif(condition=is_cicd(), reason="Skipping tests that listen on a network port for CI/CD")
    @mock.patch("webbrowser.open", mocked_browser_authserver)
    def test_get_authcode_with_browser_and_listener(self):
        under_test = AuthorizationApiClient(authorization_uri=TEST_API_ENDPOINT)

        # Cover both default and override scope paths
        # TODO: a better test would be to catch the URL that is constructed
        #       and verify it.  A real verification of that URL amounts to
        #       implementing a partial authorization server.
        authcode = under_test.authcode_from_pkce_auth_request_with_browser_and_callback_listener(
            client_id=TEST_CLIENT_ID,
            redirect_uri=self.callback_uri,
            requested_scopes=None,
            requested_audiences=None,
            pkce_code_challenge=self.pkce_challenge,
            extra=None,
        )
        self.assertEqual(MOCK_AUTHCODE, authcode)

        authcode = under_test.authcode_from_pkce_auth_request_with_browser_and_callback_listener(
            client_id=TEST_CLIENT_ID,
            redirect_uri=self.callback_uri,
            requested_scopes=TEST_REQUESTED_SCOPES,
            requested_audiences=TEST_REQUESTED_AUDIENCES,
            pkce_code_challenge=self.pkce_challenge,
            extra=None,
        )
        self.assertEqual(MOCK_AUTHCODE, authcode)

    @mock.patch("http.server.HTTPServer")
    @mock.patch("planet_auth.oidc.api_clients.authorization_api_client._parse_authcode_from_callback")
    @mock.patch("webbrowser.open")
    def test_get_authcode_with_browser_and_listener_unsupported_callback_host(self, mock1, mock2, mock3):
        under_test = AuthorizationApiClient(authorization_uri=TEST_API_ENDPOINT)

        # Only localhost callbacks should be supported by current code
        valid_callback_1 = "http://localhost:{}/test".format(self.callback_port)
        valid_callback_2 = "http://127.0.0.1:{}/test".format(self.callback_port)
        invalid_callback = "https://test.planet.com:443"

        under_test.authcode_from_pkce_auth_request_with_browser_and_callback_listener(
            client_id=TEST_CLIENT_ID,
            redirect_uri=valid_callback_1,
            requested_scopes=None,
            requested_audiences=None,
            pkce_code_challenge=self.pkce_challenge,
            extra=None,
        )

        under_test.authcode_from_pkce_auth_request_with_browser_and_callback_listener(
            client_id=TEST_CLIENT_ID,
            redirect_uri=valid_callback_2,
            requested_scopes=None,
            requested_audiences=None,
            pkce_code_challenge=self.pkce_challenge,
            extra=None,
        )

        with self.assertRaises(AuthorizationApiException):
            under_test.authcode_from_pkce_auth_request_with_browser_and_callback_listener(
                client_id=TEST_CLIENT_ID,
                redirect_uri=invalid_callback,
                requested_scopes=None,
                requested_audiences=None,
                pkce_code_challenge=self.pkce_challenge,
                extra=None,
            )

    @mock.patch("http.server.HTTPServer", spec=http.server.HTTPServer)
    @mock.patch("webbrowser.open")
    def test_get_authcode_with_browser_and_listener_unknown_callback_failure(self, mock1, mock2):
        under_test = AuthorizationApiClient(authorization_uri=TEST_API_ENDPOINT)
        with self.assertRaises(AuthorizationApiException):
            under_test.authcode_from_pkce_auth_request_with_browser_and_callback_listener(
                client_id=TEST_CLIENT_ID,
                redirect_uri=self.callback_uri,
                requested_scopes=None,
                requested_audiences=None,
                pkce_code_challenge=self.pkce_challenge,
                extra=None,
            )

    @mock.patch("getpass.getpass", mocked_get_password)
    def test_get_authcode_without_browser_and_listener(self):
        under_test = AuthorizationApiClient(authorization_uri=TEST_API_ENDPOINT)

        # Cover both default and override scope paths
        # TODO: a better test would be to catch the URL that is constructed
        #       and verify it.  A real verification of that URL amounts to
        #       implementing a partial authorization server.
        authcode = under_test.authcode_from_pkce_auth_request_with_tty_input(
            client_id=TEST_CLIENT_ID,
            redirect_uri=self.callback_uri,
            requested_scopes=None,
            requested_audiences=None,
            pkce_code_challenge=self.pkce_challenge,
            extra=None,
        )
        self.assertEqual(MOCK_AUTHCODE, authcode)

        authcode = under_test.authcode_from_pkce_auth_request_with_tty_input(
            client_id=TEST_CLIENT_ID,
            redirect_uri=self.callback_uri,
            requested_scopes=TEST_REQUESTED_SCOPES,
            requested_audiences=TEST_REQUESTED_AUDIENCES,
            pkce_code_challenge=self.pkce_challenge,
            extra=None,
        )
        self.assertEqual(MOCK_AUTHCODE, authcode)

    def test_callback_acknowledgement_config(self):
        # WARNING
        # this test is a bit clumsy and whitebox, but the blackbox way to get this data out
        # is to have HTTPServer serve it.
        under_test = AuthorizationApiClient(
            authorization_uri=TEST_API_ENDPOINT, authorization_callback_acknowledgement_response_body="literal"
        )
        self.assertEqual(under_test._authorization_callback_acknowledgement_response_body, "literal")

        from planet_auth.oidc import resources as under_test_resources

        resource_str = (
            resources.files(under_test_resources).joinpath("callback_acknowledgement.html").read_text("utf-8")
        )
        under_test = AuthorizationApiClient(
            authorization_uri=TEST_API_ENDPOINT, authorization_callback_acknowledgement_response_body=""
        )
        self.assertEqual(under_test._authorization_callback_acknowledgement_response_body, resource_str)

        under_test = AuthorizationApiClient(
            authorization_uri=TEST_API_ENDPOINT, authorization_callback_acknowledgement_response_body=None
        )
        self.assertEqual(under_test._authorization_callback_acknowledgement_response_body, resource_str)

        under_test = AuthorizationApiClient(authorization_uri=TEST_API_ENDPOINT)
        self.assertEqual(under_test._authorization_callback_acknowledgement_response_body, resource_str)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import datetime
import json
import os
import pathlib
import shutil
import tempfile
import freezegun
import unittest
import unittest.mock
import pytest

from planet_auth.credential import Credential
from planet_auth.static_api_key.request_authenticator import FileBackedApiKey
from planet_auth.storage_utils import (
    FileBackedJsonObjectException,
    InvalidDataException,
    FileBackedJsonObject,
)

from tests.test_planet_auth.unit.auth.util import MockObjectStorageProvider, MockStorageObjectNotFound
from tests.test_planet_auth.util import tdata_resource_file_path


# class MockFileBackedEntity(FileBackedJsonObject):
#     def __init__(self, data=None, file_path=None):
#         super().__init__(data=data, file_path=file_path)


class TestEntity(FileBackedJsonObject):
    def __init__(self, data=None, file_path=None):
        super().__init__(data=data, file_path=file_path)

    def check_data(self, data):
        super().check_data(data)
        # key 1 is mandatory
        if data.get("test_key_1") == "required_value_1" or data.get("test_key_1") == "required_value_2":
            pass
        else:
            raise InvalidDataException(
                "test entity requires that 'test_key_1' be 'required_value_1' or 'required_value_2"
            )

        # key 2 is optional
        if data.get("test_key_2"):
            if data.get("test_key_2") == "required_value_1" or data.get("test_key_2") == "required_value_2":
                return
            else:
                raise InvalidDataException(
                    "test entity requires that 'test_key_2' be 'required_value_1' or 'required_value_2"
                )


class TestFileBackedJsonObjectException(unittest.TestCase):
    def test_str_without_filepath(self):
        under_test = FileBackedJsonObjectException(message="test message")
        self.assertEqual("test message", f"{under_test}")

    def test_str_with_filepath(self):
        under_test = FileBackedJsonObjectException(
            message="test message", file_path=pathlib.Path("/some/unit-test/file")
        )
        self.assertEqual("test message (File: /some/unit-test/file)", f"{under_test}")


class TestFileBackedJsonObject(unittest.TestCase):
    # "FileBackedJsonObject" was originally written for the Credential
    # base class. From there, it's use expanded.  But, that remains a
    # strong influence the priority given to base functionality, and is
    # why for many test cases we simply use the Credential base class.

    def setUp(self):
        # The interactions of freezegun and the filesystem mtimes have been... quirky.
        # This seems to help.
        os.environ["TZ"] = "UTC"

    def test_set_data_asserts_valid(self):
        under_test = Credential(data=None, file_path=None)
        with self.assertRaises(FileBackedJsonObjectException):
            under_test.set_data(None)

        under_test.set_data({})

    def test_update_data_asserts_valid(self):
        under_test = TestEntity(data=None, file_path=None)
        with self.assertRaises(InvalidDataException):
            under_test.update_data(None)

        under_test.update_data({"test_key_1": "required_value_1"})
        self.assertEqual(under_test.data(), {"test_key_1": "required_value_1"})

        with self.assertRaises(InvalidDataException):
            under_test.update_data({"test_key_1": "invalid_value"})
        self.assertEqual(under_test.data(), {"test_key_1": "required_value_1"})

        under_test.update_data({"test_key_1": "required_value_2"})
        self.assertEqual(under_test.data(), {"test_key_1": "required_value_2"})

        # The update should not only be valid, but sparse.
        with self.assertRaises(InvalidDataException):
            under_test.update_data({"test_key_2": "invalid_value"})
        self.assertEqual(under_test.data(), {"test_key_1": "required_value_2"})

        under_test.update_data({"test_key_2": "required_value_1"})
        self.assertEqual(under_test.data(), {"test_key_1": "required_value_2", "test_key_2": "required_value_1"})

    def test_load_and_failed_reload(self):
        under_test = Credential(data=None, file_path=None)

        # Load fails where there is no file
        # This is no longer true.  We allow in memory use cases.
        # with self.assertRaises(FileBackedJsonObjectException):
        #     under_test.load()

        # Load works when we have a valid file
        under_test.set_path(tdata_resource_file_path("keys/base_test_credential.json"))
        under_test.load()
        self.assertEqual({"test_key": "test_value"}, under_test.data())

        # A subsequent failed load should throw, but leave the data unchanged.
        under_test.set_path(tdata_resource_file_path("keys/FILE_DOES_NOT_EXIST.json"))
        with self.assertRaises(FileNotFoundError):
            under_test.load()

        self.assertEqual({"test_key": "test_value"}, under_test.data())

    def test_load_file_not_found(self):
        under_test = Credential(data=None, file_path=tdata_resource_file_path("keys/FILE_DOES_NOT_EXIST.json"))
        with self.assertRaises(FileNotFoundError):
            # In normal operations, we would let underlying exception
            # pass to the application to handle.
            under_test.load()

    def test_load_invalid_file(self):
        # leverage a derived class we know has simple requirements we can
        # use to test the base class.
        under_test = FileBackedApiKey(
            api_key=None, api_key_file=tdata_resource_file_path("keys/invalid_test_credential.json")
        )

        with self.assertRaises(InvalidDataException):
            under_test.load()

        self.assertIsNone(under_test.data())

        with self.assertRaises(InvalidDataException):
            under_test.check()

    def test_lazy_load(self):
        # If data is not set, it should be loaded from the path, but not until
        # the data is requested.
        under_test = Credential(data=None, file_path=tdata_resource_file_path("keys/base_test_credential.json"))
        self.assertIsNone(under_test.data())
        under_test.lazy_load()
        self.assertEqual({"test_key": "test_value"}, under_test.data())

        # if the path is invalid, it should error.
        under_test = Credential(data=None, file_path=tdata_resource_file_path("keys/FILE_DOES_NOT_EXIST.json"))
        with self.assertRaises(FileNotFoundError):
            under_test.lazy_load()

        # We now permit in memory use to work.
        # under_test = Credential(data=None, file_path=None)
        # with self.assertRaises(FileBackedJsonObjectException):
        #     under_test.lazy_load()

        # Should be fine if data is set, and a lazy_load() is tried with no
        # path set or an invalid path set, no load should be performed and the
        # data should be unchanged.
        under_test = Credential(data={"ctor_key": "ctor_value"}, file_path=None)
        under_test.lazy_load()
        self.assertEqual({"ctor_key": "ctor_value"}, under_test.data())

        under_test = Credential(
            data={"ctor_key": "ctor_value"}, file_path=tdata_resource_file_path("keys/base_test_credential.json")
        )
        under_test.lazy_load()
        self.assertEqual({"ctor_key": "ctor_value"}, under_test.data())

        under_test = Credential(
            data={"ctor_key": "ctor_value"}, file_path=tdata_resource_file_path("keys/FILE_DOES_NOT_EXIST.json")
        )
        under_test.lazy_load()
        self.assertEqual({"ctor_key": "ctor_value"}, under_test.data())

    def test_lazy_reload_initial_load_behavior(self):
        # Behaves like lazy load when there is no data:
        # If data is not set, it should be loaded from the path, but not until
        # the data is asked for.
        under_test = Credential(data=None, file_path=tdata_resource_file_path("keys/base_test_credential.json"))
        self.assertIsNone(under_test.data())
        under_test.lazy_reload()
        self.assertEqual({"test_key": "test_value"}, under_test.data())

        # if the path is invalid, it should error.
        under_test = Credential(data=None, file_path=tdata_resource_file_path("keys/FILE_DOES_NOT_EXIST.json"))
        with self.assertRaises(FileNotFoundError):
            under_test.lazy_reload()

        # We now permit in memory use to work.
        # under_test = Credential(data=None, file_path=None)
        # with self.assertRaises(FileBackedJsonObjectException):
        #     under_test.lazy_reload()

        # Behaves like lazy_load when there is data but no file - contiues
        # with in memory data
        under_test = Credential(data={"ctor_key": "ctor_value"}, file_path=None)
        under_test.lazy_reload()
        self.assertEqual({"ctor_key": "ctor_value"}, under_test.data())

        # But, if the file is set, problems with the file are an error, which
        # differs from lazy_load()
        under_test = Credential(
            data={"ctor_key": "ctor_value"}, file_path=tdata_resource_file_path("keys/FILE_DOES_NOT_EXIST.json")
        )
        with self.assertRaises(FileNotFoundError):
            under_test.lazy_reload()

    @freezegun.freeze_time(as_kwarg="frozen_time")
    def test_lazy_reload_reload_behavior(self, frozen_time):
        tmp_dir = tempfile.TemporaryDirectory()
        test_path = pathlib.Path(tmp_dir.name) / "lazy_reload_test.json"
        shutil.copyfile(tdata_resource_file_path("keys/base_test_credential.json"), test_path)

        # Freezegun doesn't seem to extend to file system recorded times,
        # so monkey-patch that throughout the test, too.
        # t0 - test prep
        t0 = datetime.datetime.now(tz=datetime.timezone.utc)
        os.utime(test_path, (t0.timestamp(), t0.timestamp()))

        # t1 - object under test created
        frozen_time.tick(2)
        under_test = Credential(data=None, file_path=None)

        # test that it doesn't load until asked for
        under_test.set_path(test_path)
        self.assertIsNone(under_test.data())
        test_key_value = under_test.lazy_get("test_key")
        self.assertEqual("test_value", test_key_value)

        # t2 - update the backing file via sideband.
        # Check that changing the file DOES trigger a reload.
        # (We use a separate instance of Credential as a convenient writer to the test file.)
        new_test_data = copy.deepcopy(under_test.data())
        new_test_data["test_key"] = "new_data"
        new_credential = Credential(data=new_test_data, file_path=test_path)
        t2 = frozen_time.tick(2)
        new_credential.save()
        os.utime(test_path, (t2.timestamp(), t2.timestamp()))

        # t3 - reload the time sometime later.
        # Check to make sure we now have the new data, and that the load time was updated
        t3 = frozen_time.tick(2)
        under_test.lazy_reload()
        test_key_value = under_test.lazy_get("test_key")
        self.assertEqual("new_data", test_key_value)
        self.assertEqual(int(t3.timestamp()), under_test._load_time)

        # t4 - lazy reload the file sometime later.
        # This should NOT trigger a reload, since the file has not changed.
        old_load_time = under_test._load_time
        frozen_time.tick(2)
        under_test.lazy_reload()
        new_load_time = under_test._load_time
        self.assertEqual(old_load_time, new_load_time)

    def test_save(self):
        tmp_dir = tempfile.TemporaryDirectory()
        test_path = pathlib.Path(tmp_dir.name) / "save_test.json"
        test_data = {"some_key": "some_data"}

        # invalid data refuses to save
        under_test = Credential(data=None, file_path=test_path)
        with self.assertRaises(FileBackedJsonObjectException):
            under_test.save()

        # Path must be set - This isn't true anymore. We now allow in memory use cases.
        # under_test = Credential(data=test_data, file_path=None)
        # with self.assertRaises(FileBackedJsonObjectException):
        #     under_test.save()

        # Validate data saved correctly, and can be reconstituted in an
        # equivalent credential object.
        under_test = Credential(data=test_data, file_path=test_path)
        under_test.save()
        test_reader = Credential(data=None, file_path=test_path)
        test_reader.load()
        self.assertEqual(test_data, test_reader.data())

    def test_getters_setters(self):
        test_path = pathlib.Path("/test/test_credential.json")
        test_data = {"some_key": "some_data"}

        under_test = Credential(data=None, file_path=None)
        self.assertIsNone(under_test.data())
        self.assertIsNone(under_test.path())
        under_test.set_path(test_path)
        under_test.set_data(test_data)
        self.assertEqual(test_data, under_test.data())
        self.assertEqual(test_path, under_test.path())

    @pytest.mark.skip("No test for SOPS encryption at this time")
    def test_sops_read(self):
        pass

    @pytest.mark.skip("No test for SOPS encryption at this time")
    def test_sops_write(self):
        pass

    def test_pretty_json(self):
        def _custom_json_class_dumper(obj):
            try:
                return obj.__json_pretty_dumps__()
            except Exception:
                return obj

        def pretty_obj_str(obj):
            return json.dumps(obj, indent=0, sort_keys=True, default=_custom_json_class_dumper)

        test_data = {
            "data_1": "some_data_1",
            "data_2": None,
            "data_3": {
                "data_3_1": "some_data_3_1",
                "data_3_2": None,
            },
        }

        # In memory pretty dump
        under_test = Credential(data=test_data, file_path=None)
        pretty_str = pretty_obj_str(under_test)
        self.assertEqual('{\n"data_1": "some_data_1",\n"data_3": {\n"data_3_1": "some_data_3_1"\n}\n}', pretty_str)

        # file backed pretty dump
        under_test = Credential(data=test_data, file_path=pathlib.Path("/unit/test/dummy.json"))
        pretty_str = pretty_obj_str(under_test)
        self.assertEqual(
            '{\n"_file_path": "/unit/test/dummy.json",\n"data_1": "some_data_1",\n"data_3": {\n"data_3_1": "some_data_3_1"\n}\n}',
            pretty_str,
        )


class TestFileBackedJsonObjectCustomStorage(unittest.TestCase):
    def setUp(self):
        # Note: FileBackedJsonObject binds an instance to a path, but the storage
        #   provider is _not_ bound by that cardinality.  Multiple FileBackedJsonObject
        #   instances can use the same instances of a storage provider.
        self.utest_path1 = pathlib.Path("/utest/custom_storage/path_1")
        self.utest_path2 = pathlib.Path("/utest/custom_storage/path_2")
        self.utest_path3 = pathlib.Path("/utest/custom_storage/path_3")
        self.initial_mock_storage_1 = {self.utest_path1: {"field_1": "stored_data_1"}}
        self.initial_mock_storage_2 = {}
        self.wrapped_storage_1 = unittest.mock.Mock(
            wraps=MockObjectStorageProvider(initial_mock_storage=self.initial_mock_storage_1)
        )
        self.wrapped_storage_2 = unittest.mock.Mock(
            wraps=MockObjectStorageProvider(initial_mock_storage=self.initial_mock_storage_2)
        )

    def test_custom_load_1(self):
        under_test = FileBackedJsonObject(
            data=None, file_path=self.utest_path1, storage_provider=self.wrapped_storage_1
        )
        self.assertIsNone(under_test.data())
        under_test.lazy_load()
        self.assertEqual(self.wrapped_storage_1.load_obj.call_count, 1)
        self.assertEqual(under_test.data(), self.initial_mock_storage_1[self.utest_path1])

    def test_custom_load_2(self):
        under_test = FileBackedJsonObject(
            data={"field_1": "testcase_initial_data_1"},
            file_path=self.utest_path1,
            storage_provider=self.wrapped_storage_1,
        )
        self.assertEqual(under_test.data(), {"field_1": "testcase_initial_data_1"})
        under_test.load()  # storage data newer than initial data.
        self.assertEqual(self.wrapped_storage_1.load_obj.call_count, 1)
        self.assertEqual(under_test.data(), self.initial_mock_storage_1[self.utest_path1])

    def test_custom_load_no_path_1(self):
        under_test = FileBackedJsonObject(data=None, file_path=None, storage_provider=self.wrapped_storage_1)
        under_test.load()
        self.assertIsNone(under_test.data())
        self.assertEqual(self.wrapped_storage_1.load_obj.call_count, 0)

    def test_custom_load_no_path_2(self):
        under_test = FileBackedJsonObject(
            data={"field_1": "testcase_initial_data_1"}, file_path=None, storage_provider=self.wrapped_storage_1
        )
        under_test.load()
        self.assertEqual(under_test.data(), {"field_1": "testcase_initial_data_1"})
        self.assertEqual(self.wrapped_storage_1.load_obj.call_count, 0)

    def test_custom_load_update_path(self):
        under_test = FileBackedJsonObject(
            data=None, file_path=self.utest_path1, storage_provider=self.wrapped_storage_1
        )
        under_test.load()
        under_test.set_path(self.utest_path3)
        under_test.save()

        # Peek into mock to see that the changes we intended happened.
        # Nothing changed our mock storage at path1, so it should be the same.
        self.assertEqual(
            self.wrapped_storage_1._peek()[self.utest_path1], self.wrapped_storage_1._peek()[self.utest_path3]
        )

    def test_custom_load_not_in_storage(self):
        under_test = FileBackedJsonObject(
            data=None, file_path=self.utest_path3, storage_provider=self.wrapped_storage_2
        )
        with self.assertRaises(MockStorageObjectNotFound):
            under_test.load()

    def test_update_storage_provider_save(self):
        under_test = FileBackedJsonObject(
            data=None, file_path=self.utest_path1, storage_provider=self.wrapped_storage_1
        )
        under_test.load()
        orig_loaded_data = copy.deepcopy(under_test.data())
        under_test.set_path(self.utest_path3)
        under_test.set_storage_provider(self.wrapped_storage_2)
        self.assertEqual(under_test.storage_provider(), self.wrapped_storage_2)
        under_test.save()

        # a new object in the same storage realm gets the data from the old
        # objet writing it to the new realm after updating storage params.
        # This probes that the storage provider update happened as expected.
        # It is outside the scope of this test that custom providers work
        # as expected. Storage semantics may vary between providers.
        under_test_new = FileBackedJsonObject(
            data=None, file_path=self.utest_path3, storage_provider=self.wrapped_storage_2
        )
        under_test_new.lazy_load()
        self.assertEqual(under_test_new.data(), orig_loaded_data)

    def test_custom_save_path_saves(self):
        test_data = {"field_1": "test_custom_save_path_saves data"}
        under_test = FileBackedJsonObject(
            data=test_data, file_path=self.utest_path2, storage_provider=self.wrapped_storage_1
        )
        under_test.save()
        self.assertEqual(self.wrapped_storage_1.save_obj.call_count, 1)

        under_test_new = FileBackedJsonObject(
            data=None, file_path=self.utest_path2, storage_provider=self.wrapped_storage_1
        )
        under_test_new.load()
        self.assertEqual(under_test.data(), test_data)
        self.assertEqual(under_test.data(), under_test_new.data())

    def test_custom_save_no_path_does_not_save(self):
        test_data = {"field_1": "test_custom_save_no_path_does_not_save data"}
        under_test = FileBackedJsonObject(data=test_data, file_path=None, storage_provider=self.wrapped_storage_1)
        under_test.save()
        self.assertEqual(self.wrapped_storage_1.save_obj.call_count, 0)

        under_test_new = FileBackedJsonObject(
            data=None, file_path=self.utest_path2, storage_provider=self.wrapped_storage_1
        )
        with self.assertRaises(MockStorageObjectNotFound):
            under_test_new.load()

        self.assertIsNone(under_test_new.data())

    def test_check_if_in_storage_none(self):
        under_test = FileBackedJsonObject(file_path=None, storage_provider=self.wrapped_storage_1)
        self.assertFalse(under_test.is_persisted_to_storage())

    def test_check_if_in_storage_path_exists(self):
        under_test = FileBackedJsonObject(file_path=self.utest_path1, storage_provider=self.wrapped_storage_1)
        self.assertTrue(under_test.is_persisted_to_storage())

    def test_check_if_in_storage_path_does_not_exist(self):
        under_test = FileBackedJsonObject(file_path=self.utest_path3, storage_provider=self.wrapped_storage_1)
        self.assertFalse(under_test.is_persisted_to_storage())

    def test_check_if_in_storage_path_does_not_exist_with_init_data(self):
        test_data = {"field_1": "test_custom_save_no_path_does_not_save data"}
        under_test = FileBackedJsonObject(
            data=test_data, file_path=self.utest_path3, storage_provider=self.wrapped_storage_1
        )
        self.assertFalse(under_test.is_persisted_to_storage())
        under_test.save()
        self.assertTrue(under_test.is_persisted_to_storage())



import http
import logging

import planet_auth

import re
from flask import Flask, make_response, request
from functools import wraps


#############################################################################
# Logging Configuration
#############################################################################
# In a simple environment, the global logging format will be
# applied. This example removes all formatting around the log message from
# the outer application, allowing the library's logging to be emitted with
# no further decoration.  The auth library will emit logs in a json
# format, so removing surrounding formatting insures that the json will be
# parsable:
logging.basicConfig(format="%(message)s", level=logging.DEBUG)

# Alternatively, we can tell the library to use python structured logging.
# The exact appearance of the logs in this case will be dependent on the
# configuration of the logger.  When structured loging is used, the library
# will log using the python logger's `extra` data.  In most cases,
# this would be expected to be done in conjunction with a call to
# setPyLoggerForAuthLogger() (see below).
#
#     planet_auth.setStructuredLogging()

# What logger will be used can also be controlled.  By default, the library
# will use a logger it creates internally that is global to the library.
# Applications may set the logger for the library to "None" to quite all,
# or may set it to any logger of their choice.:
#
#     planet_auth.setPyLoggerForAuthLogger(None)
#   or
#     planet_auth.setPyLoggerForAuthLogger(logging.getLogger(name="my-custom-auth-lib-logger"))


#############################################################################
# Validator Configuration
#############################################################################
# In order to perform remote token validation with the auth server, we must
# be registered with the auth server as a client ourselves, even if we never
# intend to act as a client and obtain access tokens of our own.  In this
# example, we are registered as a confidential client (one with a client
# secret) that is permitted client credentials OAuth flow and grant type.
# This is not the only possibility.
validator_auth_client_config = {
    "client_type": "oidc_client_credentials_secret",
    "auth_server": "_trusted_auth_server_issuer_",
    "audiences": ["_expected_token_audience_"],
    "client_id": "_your_client_id_",
    "client_secret": "_your_client_secret_",
}

auth_validator = planet_auth.OidcMultiIssuerValidator.from_auth_server_configs(
    trusted_auth_server_configs=[validator_auth_client_config]
)


#############################################################################
# Connecting validator to flask using decorators (your service needs may vary)
#############################################################################
# Note: This example only enforces basic authentication - that the user has
#       authenticated to a trusted auth server and presented a valid access
#       token.  In practice, services are also responsible for determining
#       whether that particular client user is allowed to make a particular
#       request.  Services may have their own database or configuration that
#       governs this. The overall service architecture may define token
#       claims that convey particular privileges. Neither of these are
#       demonstrated by this example.
@planet_auth.AuthException.recast(ValueError)
def flask_auth_worker(do_remote_check: bool):
    auth_header = request.headers.get("Authorization")
    if not auth_header:
        raise planet_auth.AuthException("Authorization header not provided")

    scheme, token = re.split(" +", auth_header.strip(), 1)
    if scheme != "Bearer":
        raise planet_auth.AuthException("Unrecognized authentication token scheme: {}".format(scheme))

    # throws auth failures
    return auth_validator.validate_access_token(token=token, do_remote_revocation_check=do_remote_check)


def flask_auth_decorator(do_remote_check: bool = False):
    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kws):
            try:
                local_validation, remote_validation = flask_auth_worker(do_remote_check)
            except planet_auth.AuthException as ae:
                # Don't leak information to unauthorized clients in the response.
                # It's fine to be more detailed in server side logs.
                logging.warning("Auth failure: {}".format(ae))
                # Return or abort, as per your application conventions.
                return make_response("UNAUTHORIZED", http.HTTPStatus.UNAUTHORIZED)
                # abort(http.HTTPStatus.UNAUTHORIZED, "UNAUTHORIZED")
            return fn(*args, **kws)

        return wrapper

    return decorator


#############################################################################
# Main flask application
#############################################################################
app = Flask(__name__)


# Order of decorators matters
@app.route("/")
@flask_auth_decorator()
def hello():
    return "Hello World!"


@app.route("/secret")
@flask_auth_decorator(do_remote_check=True)
def secret():
    return "Sensitive info. Extra checks performed for revoked access tokens."


if __name__ == "__main__":
    app.run(host="localhost", port=5001)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from planet_auth.planet_legacy.legacy_api_key import FileBackedPlanetLegacyApiKey
from planet_auth.request_authenticator import CredentialRequestAuthenticator


class PlanetLegacyRequestAuthenticator(CredentialRequestAuthenticator):
    """
    Authenticate a request using a legacy Planet API key.  Currently,
    the credential file is now also capable of saving a JWT, but
    this authenticator makes no use of it.
    """

    TOKEN_PREFIX = "api-key"

    def __init__(self, planet_legacy_credential: FileBackedPlanetLegacyApiKey, **kwargs):
        super().__init__(credential=planet_legacy_credential, token_prefix=self.TOKEN_PREFIX, **kwargs)
        self._api_key_file = planet_legacy_credential

    def pre_request_hook(self):
        self._token_body = self._api_key_file.legacy_api_key()

    def update_credential(self, new_credential):
        if not isinstance(new_credential, FileBackedPlanetLegacyApiKey):
            raise TypeError(
                f"{type(self).__name__} does not support {type(new_credential)} credentials.  Use FileBackedPlanetLegacyApiKey."
            )
        super().update_credential(new_credential)
        self._api_key_file = new_credential

    # def update_credential_data(self, new_credential_data: Dict):
    #     super().update_credential_data(new_credential_data=new_credential_data)
    #     # The super class is not changing the instance, so we don't need to update our reference
    #     # self._api_key_file = self._credential



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
import unittest
from unittest import mock

import pytest as pytest

from planet_auth.auth_exception import AuthException, InvalidTokenException
import planet_auth.logging.auth_logger

# TODO: better assert that the right thing is being logged.
from planet_auth.logging.events import AuthEvent

MOCK_JWT_BODY_CLAIMS = {
    "iss": "unit-test-jwt-issuer",
    "aud": "unit-test-jwt-audience",
    "pl_principal": "unit-test-pl_principal",
}
MOCK_JWT_HEADER_CLAIMS = {"alg": "RS256"}


class MyException1(Exception):
    def __init__(self, msg):
        super().__init__(msg)
        self.message = msg


class MyException1Sub1(MyException1):
    pass


class MyException2(Exception):
    def __init__(self, msg):
        super().__init__(msg)
        self.message = msg


class AuthLoggerTest(unittest.TestCase):
    def setUp(self) -> None:
        self.under_test = planet_auth.logging.auth_logger.getAuthLogger()
        planet_auth.logging.auth_logger.setPyLoggerForAuthLogger(logging.getLogger("AuthLoggerTest-unit-test-logger"))

    def test_auth_logger_without_jwt(self):
        self.under_test.log(level=logging.INFO, msg="test log message")

    def test_auth_logger_with_jwt(self):
        self.under_test.log(
            level=logging.INFO,
            msg="test log message",
            jwt_header_json=MOCK_JWT_HEADER_CLAIMS,
            jwt_body_json=MOCK_JWT_BODY_CLAIMS,
        )

    def test_log_auth_exception(self):
        self.under_test.log(
            level=logging.INFO, msg="test log message", exception=AuthException(message="Test Exception")
        )

    def test_log_invalid_token_exception(self):
        self.under_test.log(
            level=logging.INFO, msg="test log message", exception=InvalidTokenException(message="Test Exception")
        )
        self.under_test.log(
            level=logging.INFO,
            msg="test log message",
            exception=InvalidTokenException(message="Test Exception"),
            jwt_header_json=MOCK_JWT_HEADER_CLAIMS,
            jwt_body_json=MOCK_JWT_BODY_CLAIMS,
        )
        self.under_test.log(
            level=logging.INFO,
            msg="test log message",
            exception=InvalidTokenException(
                message="Test Exception", jwt_header=MOCK_JWT_HEADER_CLAIMS, jwt_body=MOCK_JWT_BODY_CLAIMS
            ),
        )
        self.under_test.log(
            level=logging.INFO,
            msg="test log message",
            exception=InvalidTokenException(
                message="Test Exception", jwt_header=MOCK_JWT_HEADER_CLAIMS, jwt_body=MOCK_JWT_BODY_CLAIMS
            ),
            jwt_header_json=MOCK_JWT_HEADER_CLAIMS,
            jwt_body_json=MOCK_JWT_BODY_CLAIMS,
        )

    def test_log_exception(self):
        self.under_test.log(level=logging.INFO, msg="test log message", exception=Exception("Test Exception"))

    @mock.patch("logging.Logger.log")
    def test_decorator_logs_caught_exception(self, mock_logger):
        @self.under_test.log_exception(exception_cls=MyException1)
        def raise_my_exception1():
            raise MyException1Sub1(msg="test")

        with self.assertRaises(MyException1):
            raise_my_exception1()

        self.assertEqual(mock_logger.call_count, 1)

    @mock.patch("logging.Logger.log")
    def test_decorator_does_not_log_uncaught_exception(self, mock_logger):
        @self.under_test.log_exception(exception_cls=MyException2)
        def raise_my_exception1():
            raise MyException1(msg="test")

        with self.assertRaises(MyException1):
            raise_my_exception1()

        self.assertEqual(mock_logger.call_count, 0)

    @mock.patch("logging.Logger.log")
    def test_default_pylogger_used(self, mock_logger):
        under_test = planet_auth.logging.auth_logger.getAuthLogger()
        self.assertEqual(under_test._get_py_logger(), planet_auth.logging.auth_logger._lib_global_py_logger)
        self.assertIsNotNone(under_test._get_py_logger())
        under_test.critical(msg="test log message")
        self.assertEqual(mock_logger.call_count, 1)

    @mock.patch("logging.Logger.log")
    def test_mute_logging_with_none_pylogger(self, mock_logger):
        under_test = planet_auth.logging.auth_logger.getAuthLogger()
        planet_auth.logging.auth_logger.setPyLoggerForAuthLogger(None)
        self.assertIsNone(under_test._get_py_logger())
        under_test.critical(msg="test log message")
        self.assertEqual(mock_logger.call_count, 0)

    @mock.patch("logging.Logger.log")
    @pytest.mark.skipif(
        sys.version_info < (3, 8), reason="test requires python3.8 or higher (I believe the function should be fine)"
    )
    def test_set_structured_logging_config(self, mock_logger):
        under_test = planet_auth.logging.auth_logger.getAuthLogger()
        planet_auth.logging.auth_logger.setStructuredLogging()
        under_test.critical(msg="test log message", event=AuthEvent.TRACE)
        self.assertEqual(mock_logger.call_count, 1)
        self.assertIsNotNone(mock_logger.call_args.kwargs.get("extra"))

    @mock.patch("logging.Logger.log")
    @pytest.mark.skipif(
        sys.version_info < (3, 8),
        reason="test requires python3.8 or higher (I believe the function should be fine, since it is tested in CI/CD with other python versions)",
    )
    def test_set_structured_logging_default_nesting(self, mock_logger):
        under_test = planet_auth.logging.auth_logger.getAuthLogger()
        planet_auth.logging.auth_logger.setStructuredLogging()
        under_test.critical(msg="test log message", event=AuthEvent.TRACE)

        # The thing the test asserts is that we are nesting the extra arguments in a specific
        # dict key. We are not asserting the entire contexts of the extra data.
        # (The default value of 'props' is from json_logging)
        self.assertEqual(mock_logger.call_count, 1)
        self.assertIsNotNone(mock_logger.call_args.kwargs.get("extra"))
        self.assertIsInstance(mock_logger.call_args.kwargs.get("extra"), dict)
        self.assertIsNotNone(mock_logger.call_args.kwargs.get("extra").get("props"))
        self.assertIsInstance(mock_logger.call_args.kwargs.get("extra").get("props"), dict)
        self.assertEqual(mock_logger.call_args.kwargs.get("extra").get("props").get("event"), AuthEvent.TRACE)

    @mock.patch("logging.Logger.log")
    @pytest.mark.skipif(
        sys.version_info < (3, 8),
        reason="test requires python3.8 or higher (I believe the function should be fine, since it is tested in CI/CD with other python versions)",
    )
    def test_set_structured_logging_custom_nesting(self, mock_logger):
        under_test = planet_auth.logging.auth_logger.getAuthLogger()
        planet_auth.logging.auth_logger.setStructuredLogging(nested_key="custom_nesting_key")
        under_test.critical(msg="test log message", event=AuthEvent.TRACE)

        # The thing the test asserts is that we are nesting the extra arguments in a specific
        # dict key. We are not asserting the entire contexts of the extra data.
        # (The default value of 'props' is from json_logging)
        self.assertEqual(mock_logger.call_count, 1)
        self.assertIsNotNone(mock_logger.call_args.kwargs.get("extra"))
        self.assertIsInstance(mock_logger.call_args.kwargs.get("extra"), dict)
        self.assertIsNotNone(mock_logger.call_args.kwargs.get("extra").get("custom_nesting_key"))
        self.assertIsInstance(mock_logger.call_args.kwargs.get("extra").get("custom_nesting_key"), dict)
        self.assertEqual(
            mock_logger.call_args.kwargs.get("extra").get("custom_nesting_key").get("event"), AuthEvent.TRACE
        )

    @mock.patch("logging.Logger.log")
    @pytest.mark.skipif(
        sys.version_info < (3, 8),
        reason="test requires python3.8 or higher (I believe the function should be fine, since it is tested in CI/CD with other python versions)",
    )
    def test_set_structured_logging_no_nesting(self, mock_logger):
        under_test = planet_auth.logging.auth_logger.getAuthLogger()
        planet_auth.logging.auth_logger.setStructuredLogging(nested_key=None)
        under_test.critical(msg="test log message", event=AuthEvent.TRACE)

        # The thing the test asserts is that we are nesting the extra arguments in a specific
        # dict key. We are not asserting the entire contexts of the extra data.
        # (The default value of 'props' is from json_logging)
        self.assertEqual(mock_logger.call_count, 1)
        self.assertIsNotNone(mock_logger.call_args.kwargs.get("extra"))
        self.assertIsInstance(mock_logger.call_args.kwargs.get("extra"), dict)
        self.assertIsNotNone(mock_logger.call_args.kwargs.get("extra").get("event"))
        self.assertEqual(mock_logger.call_args.kwargs.get("extra").get("event"), AuthEvent.TRACE)

    @mock.patch("logging.Logger.log")
    @pytest.mark.skipif(
        sys.version_info < (3, 8), reason="test requires python3.8 or higher (I believe the function should be fine)"
    )
    def test_set_string_logging_config(self, mock_logger):
        under_test = planet_auth.logging.auth_logger.getAuthLogger()
        planet_auth.logging.auth_logger.setStringLogging()
        under_test.critical(msg="test log message", event=AuthEvent.TRACE)
        self.assertEqual(mock_logger.call_count, 1)
        self.assertIsNone(mock_logger.call_args.kwargs.get("extra"))

    @pytest.mark.xfail(reason="Test not yet implemented.")
    def test_decorator_event_from_default(self):
        under_test = planet_auth.logging.auth_logger.getAuthLogger()

        @under_test.log_exception(default_event=AuthEvent.TOKEN_VALID)
        def raise_my_exception1():
            raise MyException1(msg="test")

        with self.assertRaises(MyException1):
            raise_my_exception1()

        # TODO: Assert that under_test.log(event... ) got the right event (== TOKEN_VALID)
        self.fail("Test not implemented")

    @pytest.mark.xfail(reason="Test not yet implemented.")
    def test_decorator_event_from_exception(self):
        under_test = planet_auth.logging.auth_logger.getAuthLogger()

        @under_test.log_exception(default_event=AuthEvent.TOKEN_VALID)
        def raise_auth_exception():
            raise AuthException(event=AuthEvent.TOKEN_INVALID, message="test")

        with self.assertRaises(AuthException):
            raise_auth_exception()

        # TODO: Assert that under_test.log(event... ) got the right event (== TOKEN_INVALID)
        self.fail("Test not implemented")

    @pytest.mark.xfail(reason="Test not yet implemented.")
    def test_decorator_event_from_override(self):
        under_test = planet_auth.logging.auth_logger.getAuthLogger()

        @under_test.log_exception(
            default_event=AuthEvent.TOKEN_VALID, override_event=AuthEvent.TOKEN_INVALID_BAD_ISSUER
        )
        def raise_auth_exception():
            raise AuthException(event=AuthEvent.TOKEN_INVALID, message="test")

        with self.assertRaises(AuthException):
            raise_auth_exception()

        # TODO: Assert that under_test.log(event... ) got the right event (== TOKEN_INVALID_BAD_ISSUER)
        self.fail("Test not implemented")



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import unittest

from planet_auth import FileBackedOidcCredential
from planet_auth.static_api_key.auth_client import (
    StaticApiKeyAuthClient,
    StaticApiKeyAuthClientConfig,
    StaticApiKeyAuthClientException,
)
from planet_auth.static_api_key.request_authenticator import FileBackedApiKeyRequestAuthenticator
from planet_auth.static_api_key.static_api_key import FileBackedApiKey
from tests.test_planet_auth.util import tdata_resource_file_path


class TestStaticApiKeyAuthClient(unittest.TestCase):
    def setUp(self) -> None:
        self.client_conf_file = tdata_resource_file_path("auth_client_configs/utest/static_api_key_with_key.json")

    def test_login_missing_api_key(self):
        under_test = StaticApiKeyAuthClient(StaticApiKeyAuthClientConfig(api_key=None))
        with self.assertRaises(StaticApiKeyAuthClientException):
            under_test.login()

    def test_login_with_api_key(self):
        under_test = StaticApiKeyAuthClient(StaticApiKeyAuthClientConfig(api_key="test_api_key"))
        under_test_credential = under_test.login()
        self.assertIsInstance(under_test_credential, FileBackedApiKey)

    def test_default_request_authenticator_type(self):
        under_test = StaticApiKeyAuthClient(StaticApiKeyAuthClientConfig())
        test_result = under_test.default_request_authenticator(credential=pathlib.Path("/test/token.json"))
        #  test_result = under_test_no_apikey_in_conf_in_memory.default_request_authenticator(credential=None)
        self.assertIsInstance(test_result, FileBackedApiKeyRequestAuthenticator)
        self.assertEqual(pathlib.Path("/test/token.json"), test_result._credential.path())

    def test_default_request_authenticator_apikey_in_config_file(self):
        under_test_config = StaticApiKeyAuthClientConfig(file_path=self.client_conf_file)
        under_test = StaticApiKeyAuthClient(under_test_config)
        test_result = under_test.default_request_authenticator(credential=pathlib.Path("/test/token.json"))

        self.assertIsInstance(test_result, FileBackedApiKeyRequestAuthenticator)
        test_result.pre_request_hook()
        self.assertEqual("utest_static_api_key_in_file", test_result._token_body)
        self.assertEqual(self.client_conf_file, test_result._credential.path())

    def test_default_request_authenticator_apikey_in_memory_config(self):
        under_test_config = StaticApiKeyAuthClientConfig(api_key="constructor_api_key")
        under_test = StaticApiKeyAuthClient(under_test_config)
        test_result = under_test.default_request_authenticator(credential=pathlib.Path("/test/token.json"))

        self.assertIsInstance(test_result, FileBackedApiKeyRequestAuthenticator)
        test_result.pre_request_hook()
        self.assertEqual("constructor_api_key", test_result._token_body)
        self.assertIsNone(test_result._credential.path())

    def test_default_request_authenticator_given_literal_credential_object_instead_of_path(self):
        under_test_config = StaticApiKeyAuthClientConfig()
        under_test = StaticApiKeyAuthClient(under_test_config)
        literal_credential = FileBackedApiKey(api_key="literal_utest_api_credential")
        test_result = under_test.default_request_authenticator(credential=literal_credential)
        test_result.pre_request_hook()
        self.assertEqual("literal_utest_api_credential", test_result._token_body)

    def test_default_request_authenticator_given_invalid_type(self):
        under_test = StaticApiKeyAuthClient(StaticApiKeyAuthClientConfig())
        literal_credential = FileBackedOidcCredential()
        with self.assertRaises(TypeError):
            under_test.default_request_authenticator(credential=literal_credential)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import unittest

from planet_auth.auth import Auth
from planet_auth.auth_client import AuthClientException, AuthClientConfig
from planet_auth.static_api_key.auth_client import StaticApiKeyAuthClient
from planet_auth.static_api_key.request_authenticator import FileBackedApiKeyRequestAuthenticator
from planet_auth.none.noop_auth import NoOpAuthClient

from tests.test_planet_auth.util import tdata_resource_file_path


class AuthTest(unittest.TestCase):
    def test_initialize_from_conffile_with_no_token_file(self):
        under_test = Auth.initialize_from_config(
            client_config=AuthClientConfig.from_file(
                tdata_resource_file_path("auth_client_configs/utest/static_api_key.json")
            ),
        )
        self.assertIsInstance(under_test.auth_client(), StaticApiKeyAuthClient)
        self.assertIsInstance(under_test.request_authenticator(), FileBackedApiKeyRequestAuthenticator)
        self.assertIsNone(under_test.token_file_path())
        self.assertIsNone(under_test.profile_name())

    def test_initialize_from_conffile_with_token_file(self):
        under_test = Auth.initialize_from_config(
            client_config=AuthClientConfig.from_file(
                tdata_resource_file_path("auth_client_configs/utest/static_api_key.json")
            ),
            token_file="/dev/null/test_token.json",
        )
        self.assertIsInstance(under_test.auth_client(), StaticApiKeyAuthClient)
        self.assertIsInstance(under_test.request_authenticator(), FileBackedApiKeyRequestAuthenticator)
        self.assertIsInstance(under_test.token_file_path(), pathlib.Path)
        self.assertEqual(pathlib.Path("/dev/null/test_token.json"), under_test.token_file_path())
        self.assertIsNone(under_test.profile_name())

    def test_initialize_from_config(self):
        under_test = Auth.initialize_from_config_dict(
            client_config={"client_type": "none"}, token_file="/dev/null/token.json"
        )
        self.assertIsInstance(under_test.auth_client(), NoOpAuthClient)
        self.assertEqual(under_test.token_file_path(), pathlib.Path("/dev/null/token.json"))
        self.assertIsNone(under_test.profile_name())

    def test_initialize_from_config_invalid_client(self):
        with self.assertRaises(AuthClientException):
            Auth.initialize_from_config_dict(
                client_config={"client_type": "invalid"}, token_file="/dev/null/token.json"
            )

    def test_initialize_from_config_none_client(self):
        with self.assertRaises(AuthClientException):
            Auth.initialize_from_config_dict(client_config=None, token_file="/dev/null/token.json")



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

from requests.models import Response
from requests.auth import AuthBase
from typing import Tuple, Optional
from unittest import mock
import freezegun

from planet_auth.oidc.api_clients.api_client import OidcApiClientException
from planet_auth.oidc.api_clients.token_api_client import TokenApiClient, TokenApiException, TokenApiTimeoutException
from planet_auth.oidc.util import create_pkce_challenge_verifier_pair

from tests.test_planet_auth.util import mock_sleep_skip, FreezeGunMockSleep

TEST_API_ENDPOINT = "https://blackhole.unittest.planet.com/api"
TEST_CLIENT_ID = "__test_client_id__"
TEST_REDIRECT_URI = "__test_redirect_uri__"
TEST_ACCESS_TOKEN = "__test_access_token__"
TEST_ID_TOKEN = "__test_id_token__"
TEST_REFRESH_TOKEN = "__test_refresh_token__"
TEST_USER_NAME = "__test_username__"
TEST_USER_PASSWORD = "__test_password__"
API_RESPONSE_VALID = {
    "token_type": "Bearer",
    "expires_in": 3600,
    "access_token": "_dummy_access_token_",
    "scope": "offline_access profile openid",
    "refresh_token": "_dummy_refresh_token_",
    "id_token": "_dummy_id_token_",
}
API_RESPONSE_INVALID = {  # it looks valid, but no "expires_in"
    "token_type": "Bearer",
    "access_token": "_dummy_access_token_",
    "scope": "offline_access profile openid",
    "refresh_token": "_dummy_refresh_token_",
    "id_token": "_dummy_id_token_",
}
API_RESPONSE_PENDING = {
    "error": "authorization_pending",
}
API_RESPONSE_SLOW_DOWN = {
    "error": "slow_down",
}
API_RESPONSE_SERVICE_SIDE_TIMEOUT = {
    "error": "expired_token",
}
API_RESPONSE_DENIED = {
    "error": "access_denied",
}


def mocked_response_ok(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(API_RESPONSE_VALID))
    return response


def mocked_response_invalid(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(API_RESPONSE_INVALID))
    return response


def mocked_response_pending(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(API_RESPONSE_PENDING))
    return response


def mocked_response_slow_down(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(API_RESPONSE_SLOW_DOWN))
    return response


def mocked_response_denied(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(API_RESPONSE_DENIED))
    return response


def mocked_response_pending_then_valid(request_url, **kwargs):
    mocked_response_pending_then_valid.counter += 1
    if mocked_response_pending_then_valid.counter == 2:
        return mocked_response_ok(request_url, **kwargs)
    else:
        return mocked_response_pending(request_url, **kwargs)


mocked_response_pending_then_valid.counter = 0


# def load_privkey():
#     return load_rsa_private_key(
#         tdata_resource_file_path('keys/keypair1_priv.test_pem'), 'password')


class TokenApiClientTest(unittest.TestCase):
    def _test_auth_enricher(self, raw_payload: dict, audience: str) -> Tuple[dict, Optional[AuthBase]]:
        return raw_payload, None

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_ok)
    def test_token_from_authcode_valid(self, mock_post):
        under_test = TokenApiClient(token_uri=TEST_API_ENDPOINT)
        pkce_code_verifier, pkce_code_challenge = create_pkce_challenge_verifier_pair()
        token_response = under_test.get_token_from_code(
            client_id=TEST_CLIENT_ID,
            redirect_uri=TEST_REDIRECT_URI,
            code="dummy_authcode",
            code_verifier=pkce_code_verifier,
        )
        self.assertEqual(API_RESPONSE_VALID, token_response)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_ok)
    def test_token_from_authcode_valid_with_auth_enricher(self, mock_post):
        under_test = TokenApiClient(token_uri=TEST_API_ENDPOINT)
        pkce_code_verifier, pkce_code_challenge = create_pkce_challenge_verifier_pair()
        token_response = under_test.get_token_from_code(
            client_id=TEST_CLIENT_ID,
            redirect_uri=TEST_REDIRECT_URI,
            code="dummy_authcode",
            code_verifier=pkce_code_verifier,
            auth_enricher=self._test_auth_enricher,
        )
        self.assertEqual(API_RESPONSE_VALID, token_response)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_ok)
    def test_token_from_client_credentials(self, mock_post):
        under_test = TokenApiClient(token_uri=TEST_API_ENDPOINT)
        # private_key = load_privkey()
        token_response = under_test.get_token_from_client_credentials(TEST_CLIENT_ID)
        self.assertEqual(API_RESPONSE_VALID, token_response)
        # Coverage complains when we don't test the custom scopes
        # branch, but the server and it's response is mock, so we
        # don't actually have anything different to check in the result.
        # All we can check is that it didn't throw. Same for audiences.
        under_test.get_token_from_client_credentials(
            TEST_CLIENT_ID,
            requested_scopes=["scope1", "scope2"],
            requested_audiences=["req_audience_1"],
        )

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_ok)
    def test_token_from_refresh_valid(self, mock_post):
        under_test = TokenApiClient(token_uri=TEST_API_ENDPOINT)
        token_response = under_test.get_token_from_refresh(TEST_CLIENT_ID, TEST_REFRESH_TOKEN)
        self.assertEqual(API_RESPONSE_VALID, token_response)
        # Coverage complains when we don't test the custom scopes
        # branch, but the server and it's response is mock, so we
        # don't actually have anything different to check in the result.
        # All we can check is that it didn't throw.
        token_response = under_test.get_token_from_refresh(
            TEST_CLIENT_ID, TEST_REFRESH_TOKEN, requested_scopes=["scope1", "scope2"]
        )

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_invalid)
    def test_invalid_response(self, mock_post):
        under_test = TokenApiClient(token_uri=TEST_API_ENDPOINT)
        pkce_code_verifier, pkce_code_challenge = create_pkce_challenge_verifier_pair()
        with self.assertRaises(TokenApiException):
            under_test.get_token_from_code(TEST_CLIENT_ID, TEST_REDIRECT_URI, "dummy_authcode", pkce_code_verifier)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_ok)
    def test_token_from_password(self, mock_post):
        under_test = TokenApiClient(token_uri=TEST_API_ENDPOINT)
        token_response = under_test.get_token_from_password(
            client_id=TEST_CLIENT_ID, username=TEST_USER_NAME, password=TEST_USER_PASSWORD
        )
        self.assertEqual(API_RESPONSE_VALID, token_response)
        # Coverage complains when we don't test the custom scopes
        # branch, but the server and it's response is mock, so we
        # don't actually have anything different to check in the result.
        # All we can check is that it didn't throw. Same for audiences.
        under_test.get_token_from_password(
            client_id=TEST_CLIENT_ID,
            username=TEST_USER_NAME,
            password=TEST_USER_PASSWORD,
            requested_scopes=["scope1", "scope2"],
            requested_audiences=["req_audience_1"],
        )

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_ok)
    def test_token_from_password_with_auth_enricher(self, mock_post):
        under_test = TokenApiClient(token_uri=TEST_API_ENDPOINT)
        token_response = under_test.get_token_from_password(
            client_id=TEST_CLIENT_ID,
            username=TEST_USER_NAME,
            password=TEST_USER_PASSWORD,
            auth_enricher=self._test_auth_enricher,
        )
        self.assertEqual(API_RESPONSE_VALID, token_response)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_ok)
    def test_token_from_device_code(self, mock_post):
        under_test = TokenApiClient(token_uri=TEST_API_ENDPOINT)
        token_response = under_test.poll_for_token_from_device_code(
            client_id=TEST_CLIENT_ID,
            device_code="__utest_mock_device_code__",
            timeout=8,
            poll_interval=1,
            auth_enricher=self._test_auth_enricher,
        )
        self.assertEqual(API_RESPONSE_VALID, token_response)

    @mock.patch("time.sleep", mock_sleep_skip)
    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_pending_then_valid)
    def test_token_from_device_code_pending_then_ok(self, mock_post):
        under_test = TokenApiClient(token_uri=TEST_API_ENDPOINT)
        token_response = under_test.poll_for_token_from_device_code(
            client_id=TEST_CLIENT_ID,
            device_code="__utest_mock_device_code__",
            timeout=8,
            poll_interval=1,
            auth_enricher=self._test_auth_enricher,
        )
        self.assertEqual(API_RESPONSE_VALID, token_response)
        self.assertEqual(mock_post.call_count, 2)

    @freezegun.freeze_time(as_kwarg="frozen_time")
    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_slow_down)
    def test_token_from_device_code_slow_down(self, mock_post, frozen_time):
        fg_mock_sleep = FreezeGunMockSleep(frozen_time)
        with mock.patch("time.sleep", side_effect=fg_mock_sleep.mock_sleep):
            under_test = TokenApiClient(token_uri=TEST_API_ENDPOINT)
            with self.assertRaises(TokenApiTimeoutException):
                under_test.poll_for_token_from_device_code(
                    client_id=TEST_CLIENT_ID,
                    device_code="__utest_mock_device_code__",
                    timeout=16,
                    poll_interval=1,
                    auth_enricher=self._test_auth_enricher,
                )

            # A bit loose of a check. We expect a slow-down of +5 seconds each time
            # the service tells us to slow down. Our mock is simple and always tells us
            # to slow, and we exit with the client provided timeout.
            # So, we expect sleeps/checks at t = 1, 6, 11, <timeout>
            self.assertEqual(mock_post.call_count, 3)

    @freezegun.freeze_time(as_kwarg="frozen_time")
    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_pending)
    def test_token_from_device_code_client_provided_timeout(self, mock_post, frozen_time):
        fg_mock_sleep = FreezeGunMockSleep(frozen_time)
        with mock.patch("time.sleep", side_effect=fg_mock_sleep.mock_sleep):
            under_test = TokenApiClient(token_uri=TEST_API_ENDPOINT)
            with self.assertRaises(TokenApiException):
                under_test.poll_for_token_from_device_code(
                    client_id=TEST_CLIENT_ID,
                    device_code="__utest_mock_device_code__",
                    timeout=4,
                    poll_interval=1,
                    auth_enricher=self._test_auth_enricher,
                )
            # a bit loose of a check?
            self.assertEqual(mock_post.call_count, 5)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_response_denied)
    def test_token_from_device_code_service_generated_error(self, mock_post):
        under_test = TokenApiClient(token_uri=TEST_API_ENDPOINT)
        with self.assertRaises(OidcApiClientException):
            under_test.poll_for_token_from_device_code(
                client_id=TEST_CLIENT_ID,
                device_code="__utest_mock_device_code__",
                timeout=4,
                poll_interval=1,
                auth_enricher=self._test_auth_enricher,
            )
        self.assertEqual(mock_post.call_count, 1)



import click
import logging
import requests

import planet_auth_utils

logging.basicConfig(level=logging.DEBUG)


# TODO: Update (look at current planet SDK)
@click.group(help="my cli main help message")
@planet_auth_utils.opt_profile
@planet_auth_utils.opt_token_file
@click.pass_context
def my_cli_main(ctx, auth_profile, token_file):
    ctx.ensure_object(dict)
    ctx.obj["AUTH"] = planet_auth_utils.PlanetAuthFactory.initialize_auth_client_context(
        auth_profile_opt=auth_profile,
        token_file_opt=token_file,
    )


@my_cli_main.command("cmd1", help="cmd1 help message")
@click.pass_context
def do_cmd1(ctx):
    auth_ctx = ctx.obj["AUTH"]
    print("Doing cmd1 with the auth profile '{}'".format(auth_ctx.profile_name()))
    result = requests.get(
        url="http://localhost:5001/",  # flask example service.
        auth=auth_ctx.request_authenticator(),
        timeout=30,
    )
    print("Status: {}".format(result.status_code))
    print("Payload:\n{}".format(result.text))


@my_cli_main.command("cmd2", help="cmd2 help message")
@click.pass_context
def do_cmd2(ctx):
    auth_ctx = ctx.obj["AUTH"]
    print("Doing cmd2 with the auth profile '{}'".format(auth_ctx.profile_name()))
    result = requests.get(
        url="http://localhost:5001/secret",  # flask example service.
        auth=auth_ctx.request_authenticator(),
        timeout=30,
    )
    print("Status: {}".format(result.status_code))
    print("Payload:\n{}".format(result.text))


my_cli_main.add_command(planet_auth_utils.cmd_plauth_embedded)

if __name__ == "__main__":
    my_cli_main()  # pylint: disable=E1120
